/*------------------------------------------------------------------------------
    File        : bomaddhandler.p
    Description : Handle parsing of Analyst BoM to create program units and attributes		
    Author(s)   : Thomas Mercer-Hursh
    Created     : Sun Jan 13 11:23:38 PST 2008
    Notes       : Set llgDebug = true for detailed trace
                  Set llgCUToo = true to create links on CU at the same time as setting them for PU
                  Set llgPUToo = true to create links on PU at the same time as setting them for Procedure
                  
  ----------------------------------------------------------------------------*/
using com.cintegrity.Lib.Exception.*.
/* ***************************  Definitions  ******************************** */
                                                              /* "Parameters" */
define variable llgDebug as logical no-undo initial false.
define variable llgCUToo as logical no-undo initial true.
define variable llgPUToo as logical no-undo initial true.
                                                               /* Package IDs */
define variable linComponentModelPackageID as integer no-undo.   
define variable linFunctionalUnitsPackageID as integer no-undo.   
define variable linCompileUnitsPackageID as integer no-undo.   
define variable linProgramUnitsPackageID as integer no-undo.   
define variable linIncludeUnitsPackageID as integer no-undo.
define variable linCompileUnitPackageID as integer no-undo.
define variable linProgramUnitPackageID as integer no-undo.
                                                                 /* Other IDs */
define variable linCurrentCompileUnitObjectID as integer no-undo.
define variable linCurrentProgramUnitObjectID as integer no-undo.
define variable linCurrentProcedureObjectID as integer no-undo.
define variable linCurrentMethodObjectID as integer no-undo.
define variable linCurrentIncludeUnitObjectID as integer no-undo.
                                                            /* Work Variables */
define variable lchCurrentCompileUnitName as character no-undo.
define variable lchCurrentProgramUnitName as character  no-undo.
define variable lchCurrentIncludeUnitName as character  no-undo.
define variable lchCurrentProcedureName as character  no-undo.
define variable lchCurrentProcedureType as character  no-undo.
define variable lchCurrentParentElementType as character  no-undo.
define variable linPos as integer no-undo.
                                                                   /* Objects */
define variable lobExceptionStack as class SimpleExceptionStack no-undo.
                                                               /* Temp Tables */
define temp-table ttCodePackage no-undo
    field chType as character
    field chPackageName as character
    field inPackageID as integer
    field inObjectID as integer
    index ixByName as primary unique chType chPackageName
    .
/* *****************************  Functions ********************************* */
function sfCompileName returns character ( 
    input ipchProgramName as character ) 
    in super. 
/* **************************  Initialization ******************************* */
procedure Initialize:
  
  define variable iinObjectID as integer    no-undo. 
  
  lobExceptionStack = new SimpleExceptionStack().
  lobExceptionStack:trace( "Start of BoM Add Handler at " + string(now) ).
  run spFindComponentModelPackage (
      output linComponentModelPackageID
      ).
  if lobExceptionStack:numExceptions( 1 ) gt 0
  then do:
    lobExceptionStack:pushException( 401, 1, "Fatal Error in com.cintegrity.ADBtoUML.code.codemain.p - aborting" ).
    return.
  end.    
                                                  /* Create Packages for Code */
  run spCreatePackage (
      input "Compile Units",
      input linComponentModelPackageID,
      output linCompileUnitsPackageID
      ). 
  run spCreatePackage (
      input "Program Units",
      input linComponentModelPackageID,
      output linProgramUnitsPackageID
      ). 
  run spCreatePackage (
      input "Include Units",
      input linComponentModelPackageID,
      output linIncludeUnitsPackageID
      ). 
                                              /* Create corresponding objects */
  run spCreateObject (
      input "Package",
      input "Compile Units",
      input linCompileUnitsPackageID,
      input 0,
      input "",
      input "",
      input "",      
      output iinObjectID
      ).
  run spCreateObject (
      input "Package",
      input "Program Units",
      input linProgramUnitsPackageID,
      input 0,
      input "",
      input "",
      input "",
      output iinObjectID
      ).
  run spCreateObject (
      input "Package",
      input "Include Units",
      input linIncludeUnitsPackageID,
      input 0,
      input "",
      input "",
      input "",
      output iinObjectID
      ).
end procedure.
/* ***************************  Main Block  ********************************* */
procedure StartElement:
    define input  parameter ipchNameSpaceURI as character  no-undo.
    define input  parameter ipchLocalName as character  no-undo.
    define input  parameter ipchElementType as character  no-undo.
    define input  parameter iphnAttributes as handle no-undo.
    
  define variable iinProgramUnitObjectID as integer no-undo.
  define variable ichProgramUnitParentName as character no-undo.
  define variable ichProgramUnitBaseName as character no-undo.
    
  define variable iinIncludeUnitPackageID as integer no-undo.
  define variable iinIncludeUnitObjectID as integer no-undo.
  define variable ichIncludeUnitParent as character no-undo.
  define variable ichIncludeUnitBaseName as character no-undo.
  define variable iinAttributeID as integer no-undo.
  define variable iinPropertyID as integer no-undo.

  define variable iinObjectID as integer no-undo.
  define variable ichStereotype as character no-undo.
  define variable ichNotes as character no-undo.
  define variable ichAlias as character no-undo.
  define variable iinTagID as integer no-undo.
  define variable ichWhereClause as character no-undo.
  define variable ichName as character no-undo.
  define variable ichType as character no-undo.
  define variable ichWrite as character no-undo.
  define variable ichRead as character no-undo.
  define variable ichPath as character no-undo.
  
  if llgDebug then lobExceptionStack:trace( "Start of " + ipchElementType ).  
  case ipchElementType:
      when "bom" then .                                              /* start */
/* .......................................................................... */
      when "sessionsuper"                           /* Session superprocedure */
        then do:
         end.
/* .......................................................................... */
      when "adder"                            /* Prog that adds session super */
        then do:
        end.
/* .......................................................................... */
      when "cu"                                               /* Compile Unit */
        then do:
          assign
              lchCurrentProgramUnitName = iphnAttributes:get-value-by-qname("file")
              lchCurrentCompileUnitName = sfCompileName( lchCurrentProgramUnitName )
              lchCurrentParentElementType = "cu"
              linPos = 0
              .
                                                 /* Create or find CU Package */
          run FindPackage (
              input "Compile Units",
              input lchCurrentCompileUnitName,
              input linCompileUnitsPackageID,
              input "oeCompileUnit",
              output linCompileUnitPackageID,
              output linCurrentCompileUnitObjectID
              ).
          if     linCurrentCompileUnitObjectID eq ?
              or linCurrentCompileUnitObjectID eq 0
          then do:
            lobExceptionStack:pushException( 426, 2, "Failed to create compile unit package = " + lchCurrentCompileUnitName ).
            return.          
          end.
                                                         /* Set default HasDA */
          if llgCUToo
          then do:
            run spCreateObjectTag (
                input linCurrentCompileUnitObjectID,
                input "HasDA",
                input "false",
                output iinTagID
                ).     
            if iinTagID eq ?
            then do:
              lobExceptionStack:pushException( 427, 2, "Failed to create compile unit HasDA tag for " + lchCurrentCompileUnitName ).
              return.          
            end.
          end.
                                            /* Check for custom program Notes */
          run spGetProgramNotes( 
              input lchCurrentProgramUnitName,
              output ichNotes
              ).     
          if ichNotes ne ""
          then do:
            run spAddPackageNotes (
                input linCompileUnitPackageID,
                input ichNotes
                ).
          end.
                                            /* Check for custom program Alias */
          run spGetProgramAlias( 
              input lchCurrentProgramUnitName,
              output ichAlias
              ).     
          if ichAlias ne ""
          then do:
            run spAddObjectAlias (
                input linCurrentCompileUnitObjectID,
                input ichAlias
                ).
          end.          
                                            /* Check for custom program paths */
          run spGetProgramPath( 
              input lchCurrentCompileUnitName,
              output ichPath
              ).     
          if ichPath ne ""
          then do:
            run spCreateObjectTag (
                input linCurrentCompileUnitObjectID,
                input "FullPath",
                input ichPath,
                output iinPropertyID
                ).
          end.            
                                       /* Create or find PU Package and Object */
          run ParsePackage (
              input lchCurrentProgramUnitName,
              output ichProgramUnitParentName,
              output ichProgramUnitBaseName
              ).
          if ichProgramUnitParentName ne ""
          then do:
            run FindPackage (
                input "Program Units",
                input ichProgramUnitParentName,
                input linProgramUnitsPackageID,
                input "",
                output linProgramUnitPackageID,
                output iinProgramUnitObjectID
                ).
            if iinProgramUnitObjectID eq ?
            then do:
              lobExceptionStack:pushException( 428, 2, "Failed to create program unit = " + ichProgramUnitParentName ).
              return.          
            end.
          end.
          else do:
            lobExceptionStack:pushException( 402, 2, "Program Unit " + lchCurrentProgramUnitName + " has no parent package" ).
            linProgramUnitPackageID = linProgramUnitsPackageID.           
          end.  
                                                       /* Create Program Unit */
          if ichProgramUnitBaseName matches "*.cls"
          then ichStereotype = "oeClass".
          else ichStereotype = "oeProgram".   
          run spCreateObject (
              input "Component",
              input lchCurrentProgramUnitName,
              input linProgramUnitPackageID,
              input 0,
              input ichStereotype,
              input ichNotes,
              input ichAlias,
              output linCurrentProgramUnitObjectID
              ).
          if linCurrentProgramUnitObjectID eq ?
          then do:
            lobExceptionStack:pushException( 429, 2, "Failed to create program unit object = " + lchCurrentProgramUnitName ).
            return.          
          end. 
                                                         /* Set default HasDA */
          run spCreateObjectTag (
              input linCurrentProgramUnitObjectID,
              input "HasDA",
              input "false",
              output iinTagID
              ).             
          if iinTagID eq ?
          then do:
            lobExceptionStack:pushException( 430, 2, "Failed to create program unit HasDA tag for " + lchCurrentProgramUnitName ).
            return.          
          end.
                                            /* Check for custom program paths */
          run spGetProgramPath( 
              input lchCurrentProgramUnitName,
              output ichPath
              ).     
          if ichPath ne ""
          then do:
            run spCreateObjectTag (
                input linCurrentProgramUnitObjectID,
                input "FullPath",
                input ichPath,
                output iinPropertyID
                ).
          end.                      
        end.
      when "procedure"                            /* Proc within compile unit */
        then do:
          assign
              lchCurrentProcedureName = iphnAttributes:get-value-by-qname("name")
              lchCurrentProcedureType = iphnAttributes:get-value-by-qname("type")
              linPos = 0
              .
          lchCurrentProcedureName = lchCurrentProgramUnitName + ": " + lchCurrentProcedureName.    
          case lchCurrentProcedureType:
            when "Program_root"
              then do:                                        /* PROGRAM ROOT */
                /* No action required */
              end.
            when "procedure"
              then do:
                run spCreateObject (
                    input "Component",
                    input lchCurrentProcedureName,
                    input linProgramUnitPackageID,
                    input linCurrentProgramUnitObjectID,
                    input "oeIPPrivate",
                    input "",
                    input "",
                    output linCurrentProcedureObjectID
                    ).          
                if linCurrentProcedureObjectID eq ?
                then do:
                  lobExceptionStack:pushException( 431, 2, "Failed to create object for procedure " + lchCurrentProcedureName ).
                  return.          
                end.
                run spCreateObjectTag (
                    input linCurrentProcedureObjectID,
                    input "HasDA",
                    input "false",
                    output iinTagID
                    ).           

              end.
            when "method"
              then do:
                run spCreateObject (
                    input "Component",
                    input lchCurrentProcedureName,
                    input linProgramUnitPackageID,
                    input linCurrentProgramUnitObjectID,
                    input "oeMethod",
                    input "",
                    input "",
                    output linCurrentMethodObjectID
                    ).          
                if linCurrentProcedureObjectID eq ?
                then do:
                  lobExceptionStack:pushException( 481, 2, "Failed to create object for method " + lchCurrentProcedureName ).
                  return.          
                end.
                run spCreateObjectTag (
                    input linCurrentProcedureObjectID,
                    input "HasDA",
                    input "false",
                    output iinTagID
                    ).           

              end.
            when "function"
              then do:
                run spCreateObject (
                    input "Component",
                    input lchCurrentProcedureName,
                    input linProgramUnitPackageID,
                    input linCurrentProgramUnitObjectID,
                    input "oeFunctionPrivate",
                    input "",
                    input "",
                    output linCurrentProcedureObjectID
                    ).                        
                run spCreateObjectTag (
                    input linCurrentProcedureObjectID,
                    input "HasDA",
                    input "false",
                    output iinTagID
                    ).           
                if iinTagID eq ?
                then do:
                  lobExceptionStack:pushException( 431, 2, "Failed to create procedure HasDA tag for " + lchCurrentProcedureName ).
                  return.          
                end.
              end.
            otherwise do:
              lobExceptionStack:pushException( 403, 2, "Program Unit " + lchCurrentProgramUnitName + " has procedure " + lchCurrentProcedureName + " with invalid type = " + lchCurrentProcedureType ).
            end.
          end case.
        end.
      when "include"                   /* reference to include file procedure */
        then do:
          assign
              lchCurrentIncludeUnitName = iphnAttributes:get-value-by-qname("file")
              linPos = 0
              .
          run ParsePackage (
              input lchCurrentIncludeUnitName,
              output ichIncludeUnitParent,
              output ichIncludeUnitBaseName
              ).
          if ichIncludeUnitParent ne ""
          then do:
            run FindPackage (
                input "Include Units",
                input ichIncludeUnitParent,
                input linIncludeUnitsPackageID,
                input "",
                output iinIncludeUnitPackageID,
                output iinIncludeUnitObjectID
                ).
            if linCurrentProcedureObjectID eq ?
            then do:
              lobExceptionStack:pushException( 432, 2, "Failed to create object for incldue " + lchCurrentIncludeUnitName ).
              return.          
            end.
          end.
          else do:
            lobExceptionStack:pushException( 404, 2, "Include Unit " + lchCurrentIncludeUnitName + " has no parent package" ).
            iinIncludeUnitPackageID = linIncludeUnitsPackageID.           
          end.  
          run FindPackage (
              input "Include Units",
              input ichIncludeUnitParent,
              input linIncludeUnitsPackageID,
              input "",
              output iinIncludeUnitPackageID,
              output iinIncludeUnitObjectID
              ).         
          if iinIncludeUnitObjectID eq ?
          then do:
            lobExceptionStack:pushException( 433, 2, "FindPackage failed for include unit parent = " + ichIncludeUnitParent ).
            return.          
          end.
          run spCreateObject (
              input "Component",
              input ichIncludeUnitBaseName,
              input iinIncludeUnitPackageID,
              input 0,
              input "oeInclude",
              input "",
              input "",
              output linCurrentIncludeUnitObjectID
              ).          
          if linCurrentIncludeUnitObjectID eq ?
          then do:
            lobExceptionStack:pushException( 434, 2, "CreatObject failed for include unit = " + ichIncludeUnitBaseName ).
            return.          
          end.
        end.
      when "caller"                        /* Program that calls compile unit */
        then do:
        end.
      when "recordphrase"                    /* where clause within procedure */
        then do:
          ichWhereClause = iphnAttributes:get-value-by-qname("text").
          case lchCurrentParentElementType:
            when "cu"
              then do:
                run spCreateObjectTag (
                    input linCurrentProgramUnitObjectID,
                    input "WhereClause",
                    input ichWhereClause,
                    output iinTagID
                    ).         
                if iinTagID eq ?
                then do:
                  lobExceptionStack:pushException( 435, 2, "CU Whereclause tag failed for = " + lchCurrentCompileUnitName ).
                  return.          
                end.
                run spCreateObjectTag (
                    input linCurrentProgramUnitObjectID,
                    input "HasDA",
                    input "true",
                    output iinTagID
                    ).
                if iinTagID eq ?
                then do:
                  lobExceptionStack:pushException( 435, 2, "CU HasDA tag failed for = " + lchCurrentCompileUnitName ).
                  return.          
                end.
                if llgPUToo 
                then do:
                  run spCreateObjectTag (
                      input linCurrentProgramUnitObjectID,
                      input "WhereClause",
                      input ichWhereClause,
                      output iinTagID
                      ).         
                  if iinTagID eq ?
                  then do:
                    lobExceptionStack:pushException( 436, 2, "PU Whereclause tag failed for = " + lchCurrentProgramUnitName ).
                    return.          
                  end.
                  run spCreateObjectTag (
                      input linCurrentProgramUnitObjectID,
                      input "HasDA",
                      input "true",
                      output iinTagID
                      ).
                  if iinTagID eq ?
                  then do:
                    lobExceptionStack:pushException( 436, 2, "PU HasDA tag failed for = " + lchCurrentProgramUnitName ).
                    return.          
                  end.
                end.
              end.
            when "procedure"
              then do:
                run spCreateObjectTag (
                    input linCurrentProcedureObjectID,
                    input "WhereClause",
                    input ichWhereClause,
                    output iinTagID
                    ).         
                if iinTagID eq ?
                then do:
                  lobExceptionStack:pushException( 437, 2, "Proc Whereclause tag failed for = " + lchCurrentProcedureName ).
                  return.          
                end.
                run spCreateObjectTag (
                    input linCurrentProcedureObjectID,
                    input "HasDA",
                    input "true",
                    output iinTagID
                    ).
                if iinTagID eq ?
                then do:
                  lobExceptionStack:pushException( 438, 2, "Proc HasDA tag failed for = " + lchCurrentProcedureName ).
                  return.          
                end.
                if llgPUToo
                then do:
                  run spCreateObjectTag (
                      input linCurrentProgramUnitObjectID,
                      input "WhereClause",
                      input ichWhereClause,
                      output iinTagID
                      ).         
                  if iinTagID eq ?
                  then do:
                    lobExceptionStack:pushException( 438, 2, "PU Whereclause tag failed for = " + lchCurrentProgramUnitName ).
                    return.          
                  end.
                  run spCreateObjectTag (
                      input linCurrentProgramUnitObjectID,
                      input "HasDA",
                      input "true",
                      output iinTagID
                      ).
                  if iinTagID eq ?
                  then do:
                    lobExceptionStack:pushException( 439, 2, "PU HasDA tag failed for = " + lchCurrentProgramUnitName ).
                    return.          
                  end.
                end.  
                if llgCUToo 
                then do:
                  run spCreateObjectTag (
                      input linCurrentProgramUnitObjectID,
                      input "WhereClause",
                      input ichWhereClause,
                      output iinTagID
                      ).         
                  if iinTagID eq ?
                  then do:
                    lobExceptionStack:pushException( 440, 2, "CU Whereclause tag failed for = " + lchCurrentCompileUnitName ).
                    return.          
                  end.
                  run spCreateObjectTag (
                      input linCurrentProgramUnitObjectID,
                      input "HasDA",
                      input "true",
                      output iinTagID
                      ).
                  if iinTagID eq ?
                  then do:
                    lobExceptionStack:pushException( 441, 2, "CU HasDA tag failed for = " + lchCurrentCompileUnitName ).
                    return.          
                  end.
                end.  
              end.
            when "include"       /* Provided in the hope of future capability */
              then do:
               lobExceptionStack:pushException( 405, 4, "Recordphase found for Include Unit " + lchCurrentIncludeUnitName + " - not yet implemented" ).
               lobExceptionStack:pushException( 406, 4, "Recordphase = " + ichWhereClause ).
              end.
            otherwise 
              do:
                lobExceptionStack:pushException( 407, 2, "Unexpected containing unit type = " + lchCurrentParentElementType + " founc for recordphrase" ).
                lobExceptionStack:pushException( 408, 4, "Recordphase = " + ichWhereClause ).                
              end.
          end case.
        end.
      when "tableref"                     /* table reference within procedure */
        then do:
          case lchCurrentParentElementType:
            when "cu"
              then do:
                run spCreateObjectTag (
                    input linCurrentProgramUnitObjectID,
                    input "HasDA",
                    input "true",
                    output iinTagID
                    ).
                if iinTagID eq ?
                then do:
                  lobExceptionStack:pushException( 441, 2, "PU HasDA tag failed for = " + lchCurrentProgramUnitName ).
                  return.          
                end.
                if llgCUToo 
                then do:
                  run spCreateObjectTag (
                      input linCurrentCompileUnitObjectID,
                      input "HasDA",
                      input "true",
                      output iinTagID
                      ).
                  if iinTagID eq ?
                  then do:
                    lobExceptionStack:pushException( 442, 2, "CU HasDA tag failed for = " + lchCurrentCompileUnitName ).
                    return.          
                  end.
                end.
              end.
            when "procedure"
              then do:
                run spCreateObjectTag (
                    input linCurrentProcedureObjectID,
                    input "HasDA",
                    input "true",
                    output iinTagID
                    ).
                if iinTagID eq ?
                then do:
                  lobExceptionStack:pushException( 443, 2, "ProcU HasDA tag failed for = " + lchCurrentProcedureName ).
                  return.          
                end.
                if llgPUToo
                then do:
                  run spCreateObjectTag (
                      input linCurrentProgramUnitObjectID,
                      input "HasDA",
                      input "true",
                      output iinTagID
                      ).
                  if iinTagID eq ?
                  then do:
                    lobExceptionStack:pushException( 444, 2, "PU HasDA tag failed for = " + lchCurrentProgramUnitName ).
                    return.          
                  end.
                end.
                if llgCUToo 
                then do:
                  run spCreateObjectTag (
                      input linCurrentCompileUnitObjectID,
                      input "HasDA",
                      input "true",
                      output iinTagID
                      ).
                  if iinTagID eq ?
                  then do:
                    lobExceptionStack:pushException( 445, 2, "CU HasDA tag failed for = " + lchCurrentCompileUnitName ).
                    return.          
                  end.
                end.  
              end.
            when "include"       /* Provided in the hope of future capability */
              then do:
               lobExceptionStack:pushException( 409, 4, "Tableref found for Include Unit " + lchCurrentIncludeUnitName + " - not yet implemented" ).
              end.
            otherwise 
              do:
                lobExceptionStack:pushException( 410, 2, "Unexpected containing unit type = " + lchCurrentParentElementType + " founc for tableref" ).
              end.
          end case.
        end.
      when "fieldref"                     /* field reference within procedure */
        then do:
          case lchCurrentParentElementType:
            when "cu"
              then do:
                run spCreateObjectTag (
                    input linCurrentProgramUnitObjectID,
                    input "HasDA",
                    input "true",
                    output iinTagID
                    ).
                if iinTagID eq ?
                then do:
                  lobExceptionStack:pushException( 446, 2, "PU HasDA tag failed for = " + lchCurrentProgramUnitName ).
                  return.          
                end.
                if llgCUToo 
                then do:
                  run spCreateObjectTag (
                      input linCurrentProgramUnitObjectID,
                      input "HasDA",
                      input "true",
                      output iinTagID
                      ).
                  if iinTagID eq ?
                  then do:
                    lobExceptionStack:pushException( 447, 2, "CU HasDA tag failed for = " + lchCurrentCompileUnitName ).
                    return.          
                  end.
                end.
              end.
            when "procedure"
              then do:
                run spCreateObjectTag (
                    input linCurrentProcedureObjectID,
                    input "HasDA",
                    input "true",
                    output iinTagID
                    ).
                if iinTagID eq ?
                then do:
                  lobExceptionStack:pushException( 448, 2, "ProcU HasDA tag failed for = " + lchCurrentProcedureName ).
                  return.          
                end.
                if llgPUToo
                then do:
                  run spCreateObjectTag (
                      input linCurrentProgramUnitObjectID,
                      input "HasDA",
                      input "true",
                      output iinTagID
                      ).
                  if iinTagID eq ?
                  then do:
                    lobExceptionStack:pushException( 449, 2, "PU HasDA tag failed for = " + lchCurrentProgramUnitName ).
                    return.          
                  end.
                end.
                if llgCUToo 
                then do:
                  run spCreateObjectTag (
                      input linCurrentProgramUnitObjectID,
                      input "HasDA",
                      input "true",
                      output iinTagID
                      ).
                  if iinTagID eq ?
                  then do:
                    lobExceptionStack:pushException( 450, 2, "CU HasDA tag failed for = " + lchCurrentCompileUnitName ).
                    return.          
                  end.
                end.  
              end.
            when "include"       /* Provided in the hope of future capability */
              then do:
               lobExceptionStack:pushException( 411, 4, "Fieldref found for Include Unit " + lchCurrentIncludeUnitName + " - not yet implemented" ).
              end.
            otherwise 
              do:
                lobExceptionStack:pushException( 412, 2, "Unexpected containing unit type = " + lchCurrentParentElementType + " founc for fieldref" ).
              end.
          end case.          
        end.
      when "newshared"                      /* new shared object in procedure */
        then do:
          assign
              ichName = iphnAttributes:get-value-by-qname("name")
              ichType = iphnAttributes:get-value-by-qname("progresstype")
              linPos = linPos + 1
              .
          case ichType:
            when "VARIABLE"
              then do:
                case lchCurrentParentElementType:
                  when "cu"
                    then do:
                      run spCreateAttribute (
                          input linCurrentProgramUnitObjectID,
                          input ichName,
                          input "oeNewShared",
                          input "",
                          input linPos,
                          input 0,
                          input "",
                          input "",
                          input "Attribute",
                          output iinAttributeID
                          ).
                      if iinAttributeID eq ?
                      then do:
                        lobExceptionStack:pushException( 451, 2, "PU NewShared Attribute failed for = " + lchCurrentProgramUnitName ).
                        return.          
                      end.
                      if llgCUToo 
                      then do:
                        run spCreateAttribute (
                            input linCurrentProgramUnitObjectID,
                            input ichName,
                            input "oeNewShared",
                            input "",
                            input linPos,
                            input 0,
                            input "",
                            input "",
                            input "Attribute",
                            output iinAttributeID
                            ).
                        if iinAttributeID eq ?
                        then do:
                          lobExceptionStack:pushException( 452, 2, "CU NewShared Attribute failed for = " + lchCurrentCompileUnitName ).
                          return.          
                        end.
                      end.
                    end.
                  when "procedure"
                    then do:
                      run spCreateAttribute (
                          input linCurrentProcedureObjectID,
                          input ichName,
                          input "oeNewShared",
                          input "",
                          input linPos,
                          input 0,
                          input "",
                          input "",
                          input "Attribute",
                          output iinAttributeID
                          ).
                      if iinAttributeID eq ?
                      then do:
                        lobExceptionStack:pushException( 453, 2, "ProcU NewShared Attribute failed for = " + lchCurrentProcedureName ).
                        return.          
                      end.
                      if llgPUToo
                      then do:
                        run spCreateAttribute (
                            input linCurrentProgramUnitObjectID,
                            input ichName,
                            input "oeNewShared",
                            input "",
                            input linPos,
                            input 0,
                            input "",
                            input "",
                            input "Attribute",
                            output iinAttributeID
                            ).
                        if iinAttributeID eq ?
                        then do:
                          lobExceptionStack:pushException( 454, 2, "PU NewShared Attribute failed for = " + lchCurrentProgramUnitName ).
                          return.          
                        end.
                      end.
                      if llgCUToo 
                      then do:
                        run spCreateAttribute (
                            input linCurrentCompileUnitObjectID,
                            input ichName,
                            input "oeNewShared",
                            input "",
                            input linPos,
                            input 0,
                            input "",
                            input "",
                            input "Attribute",
                            output iinAttributeID
                            ).
                        if iinAttributeID eq ?
                        then do:
                          lobExceptionStack:pushException( 455, 2, "CU NewShared Attribute failed for = " + lchCurrentCompileUnitName ).
                          return.          
                        end.
                      end.  
                    end.
                  when "include" /* Provided in the hope of future capability */
                    then do:
                     lobExceptionStack:pushException( 413, 4, "Newshared found for Include Unit " + lchCurrentIncludeUnitName + " - not yet implemented" ).
                    end.
                  otherwise 
                    do:
                      lobExceptionStack:pushException( 414, 2, "Unexpected containing unit type = " + lchCurrentParentElementType + " founc for newshared" ).
                    end.
                end case. 
              end.
            when "BUFFER"
              then do:
                                                       /* Not yet implemented */
              end.         
            when "WORK-TABLE"
              then do:
                                                       /* Not yet implemented */
              end.   
            when "TEMP-TABLE"
              then do:
                                                       /* Not yet implemented */
              end.   
            when "QUERY"
              then do:
                                                       /* Not yet implemented */
              end.   
            when "FRAME"
              then do:
                                                       /* Not yet implemented */
              end.   
            when "STREAM"
              then do:
                                                       /* Not yet implemented */
              end.   
            otherwise 
              do:
                lobExceptionStack:pushException( 415, 2, "Unexpected newshared type = " + ichType ).
              end.
          end case.       
        end.
      when "sharedref"           /* reference to shared variable in procedure */
        then do:
                                      /* Read and Write not currently handled */
          assign
              ichName = iphnAttributes:get-value-by-qname("name")
              ichType = iphnAttributes:get-value-by-qname("progresstype")
              linPos = linPos + 1
              ichWrite = iphnAttributes:get-value-by-qname("write").          
              ichRead = iphnAttributes:get-value-by-qname("read").          
              .
          if ichWrite = ? then ichWrite = "false".
          if ichRead = ? then ichRead = "false".
          case ichType:
            when "VARIABLE"
              then do:
                case lchCurrentParentElementType:
                  when "cu"
                    then do:
                      run spCreateAttribute (
                          input linCurrentProgramUnitObjectID,
                          input ichName,
                          input "oeSharedRef",
                          input "",
                          input linPos,
                          input 0,
                          input "",
                          input "",
                          input "Attribute",
                          output iinAttributeID
                          ).
                      if iinAttributeID eq ?
                      then do:
                        lobExceptionStack:pushException( 456, 2, "PU NewShared Attribute failed for = " + lchCurrentProgramUnitName ).
                        return.          
                      end.
                      if llgCUToo 
                      then do:
                        run spCreateAttribute (
                            input linCurrentCompileUnitObjectID,
                            input ichName,
                            input "oeSharedRef",
                            input "",
                            input linPos,
                            input 0,
                            input "",
                            input "",
                            input "Attribute",
                            output iinAttributeID
                            ).
                        if iinAttributeID eq ?
                        then do:
                          lobExceptionStack:pushException( 457, 2, "CU NewShared Attribute failed for = " + lchCurrentCompileUnitName ).
                          return.          
                        end.
                      end.
                    end.
                  when "procedure"
                    then do:
                      run spCreateAttribute (
                          input linCurrentProcedureObjectID,
                          input ichName,
                          input "oeSharedRef",
                          input "",
                          input linPos,
                          input 0,
                          input "",
                          input "",
                          input "Attribute",
                          output iinAttributeID
                          ).
                      if iinAttributeID eq ?
                      then do:
                        lobExceptionStack:pushException( 457, 2, "ProcU NewShared Attribute failed for = " + lchCurrentProcedureName ).
                        return.          
                      end.
                      if llgPUToo
                      then do:
                        run spCreateAttribute (
                            input linCurrentProgramUnitObjectID,
                            input ichName,
                            input "oeSharedRef",
                            input "",
                            input linPos,
                            input 0,
                            input "",
                            input "",
                            input "Attribute",
                            output iinAttributeID
                            ).
                        if iinAttributeID eq ?
                        then do:
                          lobExceptionStack:pushException( 458, 2, "PU NewShared Attribute failed for = " + lchCurrentProgramUnitName ).
                          return.          
                        end.
                      end.
                      if llgCUToo 
                      then do:
                        run spCreateAttribute (
                            input linCurrentProgramUnitObjectID,
                            input ichName,
                            input "oeSharedRef",
                            input "",
                            input linPos,
                            input 0,
                            input "",
                            input "",
                            input "Attribute",
                            output iinAttributeID
                            ).
                        if iinAttributeID eq ?
                        then do:
                          lobExceptionStack:pushException( 459, 2, "CU NewShared Attribute failed for = " + lchCurrentCompileUnitName ).
                          return.          
                        end.
                      end.  
                    end.
                  when "include" /* Provided in the hope of future capability */
                    then do:
                     lobExceptionStack:pushException( 416, 4, "SharedRef found for Include Unit " + lchCurrentIncludeUnitName + " - not yet implemented" ).
                    end.
                  otherwise 
                    do:
                      lobExceptionStack:pushException( 417, 2, "Unexpected containing unit type = " + lchCurrentParentElementType + " founc for SharedRef" ).
                    end.
                end case. 
              end.
            otherwise 
              do:
                lobExceptionStack:pushException( 418, 2, "Unexpected Ssharedref type = " + ichType ).
              end.
          end case.       
        end.
      when "global"                                 /* global shared variable */
        then do:
          assign
              ichName = iphnAttributes:get-value-by-qname("name")
              linPos = linPos + 1
              .
          /* Not clear at this point what to do with these
          case lchCurrentParentElementType:
            when "cu"
              then do:
                run spCreateAttribute (
                    input linCurrentProgramUnitObjectID,
                    input ichName,
                    input "oeGlobal",
                    input "",
                    input linPos,
                    input 0,
                    input "",
                    input "",
                    input "Attribute",
                    output iinAttributeID
                    ).
                if iinAttributeID eq ?
                then do:
                  lobExceptionStack:pushException( 460, 2, "PU Global Attribute failed for = " + lchCurrentProgramUnitName ).
                  return.          
                end.
                if llgCUToo 
                then do:
                  run spCreateAttribute (
                      input linCurrentCompileUnitObjectID,
                      input ichName,
                      input "oeGlobal",
                      input "",
                      input linPos,
                      input 0,
                      input "",
                      input "",
                      input "Attribute",
                      output iinAttributeID
                      ).
                  if iinAttributeID eq ?
                  then do:
                    lobExceptionStack:pushException( 461, 2, "CU Global Attribute failed for = " + lchCurrentCompileUnitName ).
                    return.          
                  end.
                end.
              end.
            when "procedure"
              then do:
                run spCreateAttribute (
                    input linCurrentProcedureObjectID,
                    input ichName,
                    input "oeGlobal",
                    input "",
                    input linPos,
                    input 0,
                    input "",
                    input "",
                    input "Attribute",
                    output iinAttributeID
                    ).
                if iinAttributeID eq ?
                then do:
                  lobExceptionStack:pushException( 462, 2, "ProcU Global Attribute failed for = " + lchCurrentProcedureName ).
                  return.          
                end.
                if llgPUToo
                then do:
                  run spCreateAttribute (
                      input linCurrentProgramUnitObjectID,
                      input ichName,
                      input "oeGlobal",
                      input "",
                      input linPos,
                      input 0,
                      input "",
                      input "",
                      input "Attribute",
                      output iinAttributeID
                      ).
                  if iinAttributeID eq ?
                  then do:
                    lobExceptionStack:pushException( 463, 2, "PU Global Attribute failed for = " + lchCurrentProgramUnitName ).
                    return.          
                  end.
                end.
                if llgCUToo 
                then do:
                  run spCreateAttribute (
                      input linCurrentProgramUnitObjectID,
                      input ichName,
                      input "oeGlobal",
                      input "",
                      input linPos,
                      input 0,
                      input "",
                      input "",
                      input "Attribute",
                      output iinAttributeID
                      ).
                  if iinAttributeID eq ?
                  then do:
                    lobExceptionStack:pushException( 457, 2, "CU Global Attribute failed for = " + lchCurrentCompileUnitName ).
                    return.          
                  end.
                end.  
              end.
            when "include"       /* Provided in the hope of future capability */
              then do:
               lobExceptionStack:pushException( 419, 4, "Global found for Include Unit " + lchCurrentIncludeUnitName + " - not yet implemented" ).
              end.
            otherwise 
              do:
                lobExceptionStack:pushException( 420, 2, "Unexpected containing unit type = `" + lchCurrentParentElementType + "' found for Global = " + ichName ).
              end.
          end case. 
          Alternate trace message provided */
          /* Ignore this for now until we decide what to do with it
          lobExceptionStack:pushException( 420, 4, "Global reference `" + ichName + "' found and ignored - appropriate action unclear." ).
          */
        end.
      when "localsuper"                                          /* something */
        then do:
          /* Ignored until I find out what it is supposed to be */
        end.
      otherwise do:
          lobExceptionStack:pushException( 421, 1, "Unexpected Element Type in BoM = " + ipchElementType ).
        end.
  end.
end procedure.
/* -------------------------------------------------------------------------- */
procedure EndElement:
    define input  parameter ipchNameSpaceURI as character  no-undo.
    define input  parameter ipchLocalName as character  no-undo.
    define input  parameter ipchElementType as character  no-undo.
  
  if llgDebug then lobExceptionStack:trace( "Start of " + ipchElementType ).  
  case ipchElementType:
      when "bom" then .                                              /* start */
      when "sessionsuper"                           /* Session superprocedure */
        then do:
        end.
      when "adder"                            /* Prog that adds session super */
        then do:
        end.
      when "cu"                                               /* Compile Unit */
        then do:
          lchCurrentProgramUnitName = "".
          lchCurrentCompileUnitName = "".
          lchCurrentParentElementType = "".
        end.
      when "procedure"                            /* Proc within compile unit */
        then do:
          lchCurrentProcedureName = "".
          lchCurrentProcedureType = "".
          lchCurrentParentElementType = "cu".
        end.
      when "include"                   /* reference to include file procedure */
        then do:
        end.
      when "caller"                        /* Program that calls compile unit */
        then do:
        end.
      when "recordphrase"                    /* where clause within procedure */
        then do:
        end.
      when "tableref"                     /* table reference within procedure */
        then do:
        end.
      when "fieldref"                     /* field reference within procedure */
        then do:
        end.
      when "newshared"                    /* new shared variable in procedure */
        then do:
        end.
      when "sharedref"           /* reference to shared variable in procedure */
        then do:
        end.
      when "global"                                 /* global shared variable */
        then do:
        end.
      when "localsuper"                                          /* something */
        then do:
          /* Ignored until I find out what it is supposed to be */
        end.
      otherwise do:
          lobExceptionStack:pushException( 422, 1, "Unexpected End Element Type in BoM = " + ipchElementType ).
        end.
  end.
end procedure.
/* ************************  Helper Procedures ****************************** */
                                       /* Find/Create TT for Package for Code */
procedure FindPackage:
    define input parameter ipchType as character  no-undo.
    define input parameter ipchPackageName as character no-undo.
    define input parameter ipinParentPackageID as integer no-undo.
    define input parameter ipchStereotype as character no-undo.
    define output parameter opinPackageID as integer no-undo.
    define output parameter opinObjectID as integer no-undo.
  
  define variable ichPackageParentName as character  no-undo.
  define variable ichPackageBaseName as character no-undo.
  define variable iinLocalParentPackageID as integer no-undo.
  define variable iinLocalParentObjectID as integer no-undo.
  define variable iinLocalPackageID as integer no-undo.
  define variable iinLocalObjectID as integer no-undo.
  define variable ichNotes as character  no-undo.
  
  find ttCodePackage
      where ttCodePackage.chType eq ipchType 
      and   ttCodePackage.chPackage eq ipchPackageName 
      no-error.

  if not available( ttCodePackage )
  then do:
    run ParsePackage (
        input ipchPackageName,
        output ichPackageParentName,
        output ichPackageBaseName
        ).
    if ichPackageParentName ne ""
    then do:
      run FindPackage (
          input ipchType,
          input ichPackageParentName,
          input ipinParentPackageID,
          input "",
          output iinLocalParentPackageID,
          output iinLocalParentObjectID
          ).
    end.
    else do:
      iinLocalParentPackageID = ipinParentPackageID.
    end.
                                                            /* Create package */
    run spCreatePackage (
        input ipchPackageName,
        input iinLocalParentPackageID,
        output iinLocalPackageID
        ).
                                            /* Check for custom package Notes */
    run spGetPackageNote (
        input ipchPackageName,
        output ichNotes
        ).
    if ichNotes ne ""
    then do:
      run spAddPackageNotes (
          input iinLocalPackageID,
          input ichNotes
          ).
    end.
                                               /* Create corresponding object */
    run spCreateObject (
        input "Package",
        input ipchPackageName,
        input iinLocalPackageID,
        input 0,
        input ipchStereotype,
        input "",
        input "",
        output iinLocalObjectID
        ).    
                                                   /* Create Temp Table entry */
    create ttCodePackage.
    assign
        ttCodePackage.chType = ipchType
        ttCodePackage.chPackageName = ipchPackageName
        ttCodePackage.inPackageID = iinLocalPackageID
        ttCodePackage.inObjectID = iinLocalObjectID
        .
  end.  
  assign
      opinPackageID = ttCodePackage.inPackageID
      opinObjectID = ttCodePackage.inObjectID
      .
end procedure.
/* -------------------------------------------------------------------------- */
procedure ParsePackage:
  define input  parameter ipchPackageFullName as character  no-undo.
  define output parameter opchPackageParent as character  no-undo.
  define output parameter opchPackageBaseName as character  no-undo.
  
  define variable iinSlash as integer    no-undo.
  
  iinSlash = r-index(ipchPackageFullName,"/").
  if iinSlash eq 0
  then iinSlash = r-index(ipchPackageFullName,"~\").
  if iinSlash eq 0
  then
    assign
        opchPackageParent = ""
        opchPackageBaseName = ipchPackageFullName
        .
  else
    assign
        opchPackageParent = substring(ipchPackageFullName, 1, iinSlash - 1 )
        opchPackageBaseName = substring(ipchPackageFullName, iinSlash + 1 )
        .
end procedure.
/* ************************************************************************** */         
procedure BoMAddDestruct:
  delete object lobExceptionStack.
  lobExceptionStack = ?.
  empty temp-table ttCodePackage.
end procedure. 
/************************** End of Program ************************************/