/*------------------------------------------------------------------------------
    File        : bomalinkandler.p
    Description : Handle parsing of Analyst BoM to create links between programs and data		
    Author(s)   : Thomas Mercer-Hursh
    Created     : Sun Jan 13 11:23:38 PST 2008
    Notes       : Set llgDebug = true for detailed trace
                  Set llgCURunToo = true to create run links on CU at the same time as setting them for PU
                  Set llgPURunToo = true to create run links on PU at the same time as setting them for Procedure
                  Set llgCUDataToo = true to create data links on CU at the same time as setting them for PU
                  Set llgPUDataToo = true to create data links on PU at the same time as setting them for Procedure
                  Set llgSummaryPUSelfJoins = true to include summary run link self joins between program units
                  Set llgSummaryCUSelfJoins = true to include summary run link self joins between compile units
                  Set llgTrueRunSummary = true to include run summary even when link is PU to PU
                  Set llgTrueDataSummary = true to include data summary even when link is PU to table
                  
  ----------------------------------------------------------------------------*/
using com.cintegrity.Lib.Exception.*.
/* ***************************  Definitions  ******************************** */
                                                              /* "Parameters" */
define variable llgDebug as logical no-undo initial false.
define variable llgCURunToo as logical no-undo initial true.
define variable llgPURunToo as logical no-undo initial true.
define variable llgCUDataToo as logical no-undo initial true.
define variable llgPUDataToo as logical no-undo initial true.
define variable llgSummaryPUSelfJoins as logical no-undo initial false.
define variable llgSummaryCUSelfJoins as logical no-undo initial false.
define variable llgTrueRunSummary as logical no-undo initial true.
define variable llgTrueDataSummary as logical no-undo initial true.
                                                               /* Package IDs */
define variable linComponentModelPackageID as integer no-undo.   
define variable linFunctionalUnitsPackageID as integer no-undo.   
define variable linCompileUnitsPackageID as integer no-undo.   
define variable linProgramUnitsPackageID as integer no-undo.   
define variable linIncludeUnitsPackageID as integer no-undo.
                                                                 /* Other IDs */
define variable linCurrentCompileUnitObjectID as integer no-undo.
define variable linCurrentProgramUnitObjectID as integer no-undo.
define variable linCurrentProcedureObjectID as integer no-undo.
define variable linCurrentIncludeUnitObjectID as integer    no-undo.
define variable linCurrentSessionSuperPUID as integer  no-undo.
define variable linCurrentSessionSuperCUID as integer    no-undo.
                                                            /* Work Variables */
define variable lchCurrentSessionSuper as character  no-undo.                                                   
define variable lchCurrentCompileUnitName as character no-undo.
define variable lchCurrentProgramUnitName as character  no-undo.
define variable lchCurrentIncludeUnitName as character  no-undo.
define variable lchCurrentProcedureName as character  no-undo.
define variable lchCurrentProcedureType as character  no-undo.
define variable lchCurrentParentElementType as character  no-undo.
define variable lchCurrentSessionSuperPU as character  no-undo.
define variable lchCurrentSessionSuperCU as character  no-undo.
define variable lchCurrentProgramUnitTYpe as character no-undo.
define variable linPos as integer no-undo.
                                                                   /* Objects */
define variable lobExceptionStack as class SimpleExceptionStack no-undo.

/* *****************************  Functions ********************************* */
function sfCompileName returns character ( 
    input ipchProgramName as character ) 
    in super. 
/* **************************  Initialization ******************************* */
procedure Initialize:
  
  lobExceptionStack = new SimpleExceptionStack().
  lobExceptionStack:trace( "Start of BoM Link Handler at " + string(now) ).

end.
/* ***************************  Main Block  ********************************* */
procedure StartElement:
    define input  parameter ipchNameSpaceURI as character  no-undo.
    define input  parameter ipchLocalName as character  no-undo.
    define input  parameter ipchElementType as character  no-undo.
    define input  parameter iphnAttributes as handle no-undo.
    
  define variable iinCompileUnitPackageID as integer no-undo.
  
  define variable iinProgramUnitPackageID as integer no-undo.
  define variable iinProgramUnitObjectID as integer no-undo.
  define variable ichProgramUnitParentName as character no-undo.
  define variable ichProgramUnitBaseName as character no-undo.
    
  define variable iinIncludeUnitPackageID as integer no-undo.
  define variable iinIncludeUnitObjectID as integer no-undo.
  define variable ichIncludeUnitParent as character no-undo.
  define variable ichIncludeUnitBaseName as character no-undo.
  define variable iinAttributeID as integer no-undo.
  define variable iinAdderID as integer    no-undo.
  define variable iinConnectorID as integer    no-undo.
  define variable iinConnectorTagID as integer    no-undo.
  define variable iinAdderObjectID as integer    no-undo.
  define variable iinCallerProcedureID as integer    no-undo.
  define variable iinCallerProgramUnitID as integer    no-undo.
  define variable iinCallerCompileUnitID as integer    no-undo.
  define variable iinCallerObjectID as integer no-undo.

  define variable iinObjectID as integer no-undo.
  define variable ichStereotype as character no-undo.
  define variable ichNotes as character no-undo.
  define variable iinTagID as integer no-undo.
  define variable ichWhereClause as character no-undo.
  define variable ichName as character no-undo.
  define variable ichType as character no-undo.
  define variable ichWrite as character no-undo.
  define variable ichRead as character no-undo.
  define variable ichAdderPU as character  no-undo.
  define variable ichAdderCU as character  no-undo.
  define variable ichAdderProc as character  no-undo.
  define variable ichCallerProgramUnitName as character  no-undo.
  define variable ichCallerProcedureName as character  no-undo.
  define variable ichCallerCompileUnitName as character  no-undo.
  define variable ilgCreate as logical no-undo.
  define variable ilgDelete as logical no-undo.
  define variable ilgRead as logical no-undo.
  define variable ilgWrite as logical no-undo.
  define variable ichTableName as character  no-undo.
  define variable iinTableObjectID as integer    no-undo.
  define variable ichLockStatus as character  no-undo.
  define variable ichFieldName as character  no-undo.
  define variable ichAccessMode as character  no-undo.

  if llgDebug then lobExceptionStack:trace( "Start of " + ipchElementType ).  
  case ipchElementType:
      when "bom" then .                                              /* start */
/* .......................................................................... */
      when "sessionsuper"                           /* Session superprocedure */
        then do:
          lchCurrentSessionSuperPU = iphnAttributes:get-value-by-qname("cu").
          run spFindObject (
              input "Component",
              input "oeProgram",
              input lchCurrentSessionSuperPU,
              output linCurrentSessionSuperPUID
              ).          
          if      linCurrentSessionSuperPUID eq ?
               or linCurrentSessionSuperPUID eq 0
          then do:
            lobExceptionStack:pushException( 501, 2, "FindObject failed for = `" + lchCurrentSessionSuperPU + "' in sessionsuper" ).
            return.          
          end.
          lchCurrentSessionSuperCU = sfCompileName( lchCurrentSessionSuperPU ).
          run spFindObject (
              input "Package",
              input "oeCompileUnit",
              input lchCurrentSessionSuperCU,
              output linCurrentSessionSuperCUID
              ).                    
          if     linCurrentSessionSuperCUID eq ?
              or linCurrentSessionSuperCUID eq 0
          then do:
            lobExceptionStack:pushException( 502, 2, "Find Object failed for = `" + lchCurrentSessionSuperCU + "' in sessionsuper" ).
            return.          
          end.
        end.
/* .......................................................................... */
      when "adder"                            /* Prog that adds session super */
        then do:
          ichAdderPU = iphnAttributes:get-value-by-qname("cu").
          ichAdderCU = sfCompileName( ichAdderPU ).
          ichAdderProc = iphnAttributes:get-value-by-qname("proc").
          if ichAdderProc ne ""
          then do:
            ichAdderProc = ichAdderPU + ":" + ichAdderProc.
          end.
          
          if ichAdderProc ne ""
          then do:
            run spFindObject (
                input "Component",
                input "oeIPPrivate",
                input ichAdderProc,
                output iinAdderObjectID
                ).
            if     iinAdderObjectID eq ?
                or iinAdderObjectID eq 0
            then do:
              lobExceptionStack:pushException( 511, 2, "FindObject failed for = `" + ichAdderProc + "' in adder" ).
              return.          
            end.
            run spCreateConnector (
                input "Dependency",
                input "",
                input "",
                input "",
                input "",
                input iinAdderObjectID,
                input linCurrentSessionSuperPUID,
                input "oeAddSuper",
                output iinConnectorID 
                ).
            if     iinConnectorID eq ?
                or iinConnectorID eq 0
            then do:
              lobExceptionStack:pushException( 512, 2, "CreateConnector failed for = `" + ichAdderProc + "' in adder" ).
              return.          
            end.
            run spCreateConnectorTag (
                input iinConnectorID,
                input "IsSessionSuper",
                input "true",
                output iinConnectorTagID
                ).
            if iinConnectorTagID eq ?
            then do:
              lobExceptionStack:pushException( 513, 2, "CreateConnectorTag failed for = `" + ichAdderProc + "' in adder" ).
              return.          
            end.
          end.
          else do:
            run spFindObject (
                input "Component",
                input "oeProgram",
                input ichAdderPU,
                output iinAdderObjectID
                ).
            if     iinAdderObjectID eq ?
                or iinAdderObjectID eq 0
            then do:
              lobExceptionStack:pushException( 514, 2, "FindObject failed for = `" + ichAdderPU + "' in adder" ).
              return.          
            end.
            run spCreateConnector (
                input "Dependency",
                input "",
                input "",
                input "",
                input "",
                input iinAdderObjectID,
                input linCurrentSessionSuperPUID,
                input "oeAddSuper",
                output iinConnectorID 
                ).
            if     iinConnectorID eq ?
                or iinConnectorID eq 0
            then do:
              lobExceptionStack:pushException( 515, 2, "Create Connector failed for = " + ichAdderPU + " in adder" ).
              return.          
            end.
            run spCreateConnectorTag (
                input iinConnectorID,
                input "IsSessionSuper",
                input "true",
                output iinConnectorTagID
                ).
            if     iinConnectorTagID eq ?
                or iinConnectorTagID eq 0
            then do:
              lobExceptionStack:pushException( 516, 2, "CreateConnectTag failed for = " + ichAdderPU + " in adder" ).
              return.          
            end.
          end.
          
          if      llgPURunToo
              and ( ichAdderProc ne "" or llgTrueRunSummary ) 
          then do:
            run spFindObject (
                input "Component",
                input "oeProgram",
                input ichAdderPU,
                output iinAdderObjectID
                ).
            if     iinAdderObjectID eq ?
                or iinAdderObjectID eq 0
            then do:
              lobExceptionStack:pushException( 517, 2, "FindObject failed for = `" + ichAdderPU + "' in adder" ).
              return.          
            end.
            if     llgSummaryPUSelfJoins
                or iinAdderObjectID ne linCurrentSessionSuperPUID
            then do:
              run spCreateConnector (
                  input "Dependency",
                  input "",
                  input "",
                  input "",
                  input "",
                  input iinAdderObjectID,
                  input linCurrentSessionSuperPUID,
                  input "oeAddSuperSummary",
                  output iinConnectorID 
                  ).
              if     iinConnectorID eq ?
                  or iinConnectorID eq 0
              then do:
                lobExceptionStack:pushException( 518, 2, "Create Connector failed for = " + ichAdderPU + " in adder" ).
                return.          
              end.
              run spCreateConnectorTag (
                  input iinConnectorID,
                  input "IsSessionSuper",
                  input "true",
                  output iinConnectorTagID
                  ).
              if     iinConnectorTagID eq ?
                  or iinConnectorTagID eq 0
              then do:
                lobExceptionStack:pushException( 519, 2, "CreateConnectTag failed for = " + ichAdderPU + " in adder" ).
                return.          
              end.
            end.
          end.  
          
          if llgCURunToo
          then do:
            run spFindObject (
                input "Package",
                input "oeCompileUnit",
                input ichAdderCU,
                output iinAdderObjectID
                ).
            if     iinAdderObjectID eq ?
                or iinAdderObjectID eq 0
            then do:
              lobExceptionStack:pushException( 520, 2, "FindObject failed for = " + ichAdderCU + " in adder" ).
              return.          
            end.
            if     llgSummaryCUSelfJoins
                or iinAdderObjectID ne linCurrentSessionSuperCUID
            then do:
              run spCreateConnector (
                  input "Dependency",
                  input "",
                  input "",
                  input "",
                  input "",
                  input iinAdderObjectID,
                  input linCurrentSessionSuperCUID,
                  input "oeAddSuperSummary",
                  output iinConnectorID 
                  ).
              if     iinConnectorID eq ?
                  or iinConnectorID eq 0
              then do:
                lobExceptionStack:pushException( 521, 2, "CreateConnector failed for = " + ichAdderCU + " in adder" ).
                return.          
              end.
              run spCreateConnectorTag (
                  input iinConnectorID,
                  input "IsSessionSuper",
                  input "true",
                  output iinConnectorTagID
                  ).            
              if     iinConnectorTagID eq ?
                  or iinConnectorTagID eq 0
              then do:
                lobExceptionStack:pushException( 522, 2, "CreateConnectorTag failed for = " + ichAdderCU + " in adder" ).
                return.          
              end.
            end.
          end.
        end.
/* .......................................................................... */
      when "cu"                                               /* Compile Unit */
        then do:
          assign
              lchCurrentProgramUnitName = iphnAttributes:get-value-by-qname("file")
              lchCurrentCompileUnitName = sfCompileName( lchCurrentProgramUnitName )
              lchCurrentParentElementType = "cu"
              linPos = 0
              .
          if lchCurrentProgramUnitName matches "*.cls"
          then do:
            lchCurrentProgramUnitType = "CLASS".
            run spFindObject (
                input "Component",
                input "oeClass",
                input lchCurrentProgramUnitName,
                output linCurrentProgramUnitObjectID
                ).                  
          end.
          else do:
            lchCurrentProgramUnitType = "PROGRAM".
            run spFindObject (
                input "Component",
                input "oeProgram",
                input lchCurrentProgramUnitName,
                output linCurrentProgramUnitObjectID
                ).      
          end.    
          if     linCurrentProgramUnitObjectID eq ?
              or linCurrentProgramUnitObjectID eq 0
          then do:
            lobExceptionStack:pushException( 505, 2, "FindObject failed for = " + lchCurrentProgramUnitName + " in cu" ).
            return.          
          end.
          run spFindObject (
              input "Package",
              input "oeCompileUnit",
              input lchCurrentCompileUnitName,
              output linCurrentCompileUnitObjectID
              ).
          if     linCurrentCompileUnitObjectID eq ?
              or linCurrentCompileUnitObjectID eq 0
          then do:
            lobExceptionStack:pushException( 506, 2, "FindObject failed for = " + lchCurrentCompileUnitName + " in cu" ).
            return.          
          end.
        end.
/* .......................................................................... */
      when "procedure"                            /* Proc within compile unit */
        then do:
          assign
              lchCurrentProcedureName = iphnAttributes:get-value-by-qname("name")
              lchCurrentProcedureType = iphnAttributes:get-value-by-qname("type")
              linPos = 0
              .
          if lchCurrentProcedureType ne "Program_Root"
          then do:
            assign
                lchCurrentParentElementType = "procedure"
                lchCurrentProcedureName = lchCurrentProgramUnitName + ": " + lchCurrentProcedureName
                .
            case lchCurrentProcedureType:
              when "PROCEDURE"
                then do:
                  run spFindObject (
                      input "Component",
                      input "oeIPPrivate",
                      input lchCurrentProcedureName,
                      output linCurrentProcedureObjectID
                      ).               
                end.
              when "FUNCTION"
                then do:
                  run spFindObject (
                      input "Component",
                      input "oeFunctionPrivate",
                      input lchCurrentProcedureName,
                      output linCurrentProcedureObjectID
                      ).                              
                end.
              when "METHOD"
                then do:
                  run spFindObject (
                      input "Component",
                      input "oeMethod",
                      input lchCurrentProcedureName,
                      output linCurrentProcedureObjectID
                      ).                              
                end.
              otherwise
                do:
                  /* reported by bomaddhandler */
                end.
            end case.
            if     linCurrentProcedureObjectID eq ?
                or linCurrentProcedureObjectID eq 0
            then do:
              lobExceptionStack:pushException( 507, 2, "Find Object failed for procedure = " + lchCurrentProcedureName + " of type " + lchCurrentProcedureType + " in <procedure" ).
              return.          
            end.
          end.
      
        end.
/* .......................................................................... */
      when "include"                   /* reference to include file procedure */
        then do:
          assign
              lchCurrentIncludeUnitName = iphnAttributes:get-value-by-qname("file")
              linPos = 0
              .
        end.
/* .......................................................................... */        
      when "caller"                        /* Program that calls compile unit */
        then do:
          assign
              ichCallerProgramUnitName = iphnAttributes:get-value-by-qname("cu")
              ichCallerCompileUnitName = sfCompileName( ichCallerProgramUnitName )
              ichCallerProcedureName = iphnAttributes:get-value-by-qname("proc")
              .
          if ichCallerProcedureName ne ""
          then do:
            ichCallerProcedureName = ichCallerProgramUnitName + ": " + ichCallerProcedureName.
            run spFindObject (
                input "Component",
                input "oeIPPrivate",
                input ichCallerProcedureName,
                output iinCallerProcedureID
                ).
            if     iinCallerProcedureID eq ?
                or iinCallerProcedureID eq 0
            then do:
              run spFindObject (
                  input "Component",
                  input "oeFunctionPrivate",
                  input ichCallerProcedureName,
                  output iinCallerProcedureID
                  ).              
              if     iinCallerProcedureID eq ?
                  or iinCallerProcedureID eq 0
              then do:
                run spFindObject (
                    input "Component",
                    input "oeMethod",
                    input ichCallerProcedureName,
                    output iinCallerProcedureID
                    ).              
                if     iinCallerProcedureID eq ?
                    or iinCallerProcedureID eq 0
                then do:
                  lobExceptionStack:pushException( 531, 2, "Find Object failed for = " + ichCallerProcedureName + " in caller" ).
                  return.
                end.          
              end.
            end.  
          end.

          run spFindObject (
              input "Component",
              input "oeProgram",
              input ichCallerProgramUnitName,
              output iinCallerProgramUnitID
              ).
          if     iinCallerProgramUnitID eq ?
              or iinCallerProgramUnitID eq 0
          then do:
            lobExceptionStack:pushException( 532, 2, "Find Object failed for = " + ichCallerProgramUnitName + " in caller" ).
            return.          
          end.
          run spFindObject (
              input "Package",
              input "oeCompileUnit",
              input ichCallerCompileUnitName,
              output iinCallerCompileUnitID
              ).
          if     iinCallerCompileUnitID eq ?
              or iinCallerCompileUnitID eq 0
          then do:
            lobExceptionStack:pushException( 533, 2, "Find Object failed for = " + ichCallerCompileUnitName + " in caller" ).
            return.          
          end.

          case lchCurrentParentElementType:
            when "procedure"
              then do:
                case lchCurrentProcedureType:
                  when "PROCEDURE" then ichStereotype = "oeRunIPPrivate".
                  when "FUNCTION" then ichStereotype = "oeRunFunctionPrivate".
                  when "METHOD" then ichStereotype = "oeRunMethoc".
                  otherwise do:
                    lobExceptionStack:pushException( 535, 2, "Invalid Current Procedure Type = " + lchCurrentProcedureType + " in caller" ).
                    return.          
                  end.
                end case.      
                if ichCallerProcedureName ne ""
                then do:
                  run spCreateConnector (
                      input "ControlFlow",
                      input "",
                      input "",
                      input "",
                      input "",
                      input iinCallerProcedureID,
                      input linCurrentProcedureObjectID,
                      input ichStereotype,
                      output iinConnectorID 
                      ).
                  if     iinConnectorID eq ?
                      or iinConnectorID eq 0
                  then do:
                    lobExceptionStack:pushException( 534, 2, "CreateConnector failed for = " + ichCallerProcedureName + " in caller" ).
                    return.          
                  end.
                end.
                else do:
                  run spCreateConnector (
                      input "ControlFlow",
                      input "",
                      input "",
                      input "",
                      input "",
                      input iinCallerProgramUnitID,
                      input linCurrentProcedureObjectID,
                      input ichStereotype,
                      output iinConnectorID 
                      ).
                  if     iinConnectorID eq ?
                      or iinConnectorID eq 0
                  then do:
                    lobExceptionStack:pushException( 535, 2, "CreateConnector failed for = " + ichCallerProgramUnitName + " in caller" ).
                    return.          
                  end.
                end.                
                if      llgPURunToo
                    and ( ichCallerProcedureName ne "" or llgTrueRunSummary )
                then do:
                  if     llgSummaryPUSelfJoins
                      or iinCallerProgramUnitID ne linCurrentProgramUnitObjectID
                  then do:
                    run spCreateConnector (
                        input "ControlFlow",
                        input "",
                        input "",
                        input "",
                        input "",
                        input iinCallerProgramUnitID,
                        input linCurrentProgramUnitObjectID,
                        input "oeRunSummary",
                        output iinConnectorID 
                        ).
                    if     iinConnectorID eq ?
                        or iinConnectorID eq 0
                    then do:
                      lobExceptionStack:pushException( 536, 2, "CreateConnector failed for = " + ichCallerCompileUnitName + " in caller" ).
                      return.          
                    end.
                  end.
                end.
              end.
            when "cu"
              then do:
                case lchCurrentProgramUnitType:
                  when "PROGRAM" then ichStereotype = "oeRunProgram".
                  when "CLASS" then ichStereotype = "oeRunClass".
                  otherwise do:
                    lobExceptionStack:pushException( 537, 2, "Invalid Current Program Unit Type = " + lchCurrentProgramUnitType + " in caller" ).
                    return.          
                  end.
                end case.      
                if ichCallerProcedureName ne ""
                then do:
                  run spCreateConnector (
                      input "ControlFlow",
                      input "",
                      input "",
                      input "",
                      input "",
                      input iinCallerProcedureID,
                      input linCurrentProgramUnitObjectID,
                      input ichStereotype,
                      output iinConnectorID 
                      ).
                  if     iinConnectorID eq ?
                      or iinConnectorID eq 0
                  then do:
                    lobExceptionStack:pushException( 538, 2, "CreateConnector failed for = " + ichCallerProgramUnitName + " in caller" ).
                    return.          
                  end.
                end.
                else do:
                  run spCreateConnector (
                      input "ControlFlow",
                      input "",
                      input "",
                      input "",
                      input "",
                      input iinCallerProgramUnitID,
                      input linCurrentProgramUnitObjectID,
                      input ichStereotype,
                      output iinConnectorID 
                      ).
                  if     iinConnectorID eq ?
                      or iinConnectorID eq 0
                  then do:
                    lobExceptionStack:pushException( 539, 2, "CreateConnector failed for = " + ichCallerProgramUnitName + " in caller" ).
                    return.          
                  end.
                end.                
                if      llgPURunToo
                    and ( ichCallerProcedureName ne "" or llgTrueRunSummary )
                then do:
                  if     llgSummaryPUSelfJoins
                      or iinCallerProgramUnitID ne linCurrentProgramUnitObjectID
                  then do:
                    run spCreateConnector (
                        input "ControlFlow",
                        input "",
                        input "",
                        input "",
                        input "",
                        input iinCallerProgramUnitID,
                        input linCurrentProgramUnitObjectID,
                        input "oeRunSummary",
                        output iinConnectorID 
                        ).
                    if     iinConnectorID eq ?
                        or iinConnectorID eq 0
                    then do:
                      lobExceptionStack:pushException( 540, 2, "CreateConnector failed for = " + ichCallerProgramUnitName + " in caller" ).
                      return.          
                    end.
                  end.
                end.
              end.
            otherwise 
              do:
                lobExceptionStack:pushException( 541, 2, "Unexpected containing unit type = " + lchCurrentParentElementType + " founc for caller" ).
              end.
          end case.
          
          if llgCURunToo
          then do:
            if     llgSummaryCUSelfJoins
                or iinCallerCompileUnitID ne linCurrentCompileUnitObjectID
            then do:
              run spCreateConnector (
                  input "ControlFlow",
                  input "",
                  input "",
                  input "",
                  input "",
                  input iinCallerCompileUnitID,
                  input linCurrentCompileUnitObjectID,
                  input "oeRunSummary",
                  output iinConnectorID 
                  ).
              if     iinConnectorID eq ?
                  or iinConnectorID eq 0
              then do:
                lobExceptionStack:pushException( 542, 2, "CreateConnector failed for = " + ichCallerCompileUnitName + " in caller" ).
                return.          
              end.
            end.
          end.
        end.
/* .......................................................................... */
      when "recordphrase"                    /* where clause within procedure */
        then do:
          ichWhereClause = iphnAttributes:get-value-by-qname("text").
          ichTableName = entry( 1, ichWhereClause, ":" ).
          if index( ichWhereClause, "NO-LOCK" ) gt 0
          then ichLockStatus = "NO-LOCK".
          else if index( ichWhereClause, "EXCLUSIVE-LOCK" ) gt 0
          then ichLockStatus = "EXCLUSIVE-LOCK".
          else ichLockStatus = "SHARE-LOCK".
          
          run spFindObject (
              input "Class",
              input "table",
              input ichTableName,
              output iinTableObjectID
              ).
          if     iinTableObjectID eq ?
              or iinTableObjectID eq 0
          then do:
            lobExceptionStack:pushException( 551, 2, "FindObject failed for = " + ichTableName + " in recordphrase in " + lchCurrentProgramUnitName ).
            return.          
          end.
          
          case lchCurrentParentElementType:
            when "cu"
              then do:
                run spDataLink (
                    input linCurrentProgramUnitObjectID,
                    input iinTableObjectID,
                    input false,
                    input true,
                    input false,
                    input false,
                    input false,
                    output iinConnectorID
                    ).
                if     iinConnectorID eq ?
                    or iinConnectorID eq 0
                then do:
                  lobExceptionStack:pushException( 552, 2, "CreateConnector failed for = " + ichTableName + " in recordphrase in " + lchCurrentProgramUnitName ).
                  return.          
                end.
                run spCreateConnectorTag (
                    input iinConnectorID,
                    input "WhereClause",
                    input ichWhereClause,
                    output iinTagID
                    ).         
                if     iinTagID eq ?
                    or iinTagID eq 0
                then do:
                  lobExceptionStack:pushException( 553, 2, "CreateConnectorTag Whereas failed for = " + ichTableName + " in recordphrase in " + lchCurrentProgramUnitName ).
                  return.          
                end.
                run spCreateConnectorTag (
                    input iinConnectorID,
                    input "LockStatus",
                    input ichLockStatus,
                    output iinTagID
                    ).         
                if     iinTagID eq ?
                    or iinTagID eq 0
                then do:
                  lobExceptionStack:pushException( 554, 2, "CreateConnectorTag LockStatus failed for = " + ichTableName + " in recordphrase in " + lchCurrentProgramUnitName ).
                  return.          
                end.
              end.
            when "procedure"
              then do:
                run spDataLink (
                    input linCurrentProcedureObjectID,
                    input iinTableObjectID,
                    input false,
                    input true,
                    input false,
                    input false,
                    input false,
                    output iinConnectorID
                    ).
                if     iinConnectorID eq ?
                    or iinConnectorID eq 0
                then do:
                  lobExceptionStack:pushException( 558, 2, "CreateConnector failed for = " + ichTableName + " in recordphrase in " + lchCurrentProgramUnitName ).
                  return.          
                end.
                run spCreateConnectorTag (
                    input iinConnectorID,
                    input "WhereClause",
                    input ichWhereClause,
                    output iinTagID
                    ).         
                  if     iinTagID eq ?
                      or iinTagID eq 0
                  then do:
                    lobExceptionStack:pushException( 559, 2, "CreateConnectorTag Whereclause failed for = " + ichTableName + " in recordphrase in " + lchCurrentProgramUnitName ).
                    return.          
                  end.
                run spCreateConnectorTag (
                    input iinConnectorID,
                    input "LockStatus",
                    input ichLockStatus,
                    output iinTagID
                    ).         
                if     iinTagID eq ?
                    or iinTagID eq 0
                then do:
                  lobExceptionStack:pushException( 560, 2, "CreateConnectorTag LockStatus failed for = " + ichTableName + " in recordphrase in " + lchCurrentProgramUnitName ).
                  return.          
                end.
              end.
            when "include"       /* Provided in the hope of future capability */
              then do:
                                            /* reported on by bomaddhandler.p */
              end.
            otherwise 
              do:
                                            /* reported on by bomaddhandler.p */                
              end.
          end case.
          if llgPURunToo
          then do:
            run spDataLink (
                input linCurrentProgramUnitObjectID,
                input iinTableObjectID,
                input false,
                input true,
                input false,
                input false,
                input true,
                output iinConnectorID
                ).
            if     iinConnectorID eq ?
                or iinConnectorID eq 0
            then do:
              lobExceptionStack:pushException( 561, 2, "CreateConnector failed for = " + ichTableName + " in recordphrase in " + lchCurrentProgramUnitName ).
              return.          
            end.
            run spCreateConnectorTag (
                input iinConnectorID,
                input "WhereClause",
                input ichWhereClause,
                output iinTagID
                ).         
            if     iinTagID eq ?
                or iinTagID eq 0
            then do:
              lobExceptionStack:pushException( 562, 2, "CreateConnectorTag Whereclause failed for = " + ichTableName + " in recordphrase in " + lchCurrentProgramUnitName ).
              return.          
            end.
            run spCreateConnectorTag (
                input iinConnectorID,
                input "LockStatus",
                input ichLockStatus,
                output iinTagID
                ).         
            if     iinTagID eq ?
                or iinTagID eq 0
            then do:
              lobExceptionStack:pushException( 563, 2, "CreateConnectorTag LockStatus failed for = " + ichTableName + " in recordphrase in " + lchCurrentProgramUnitName ).
              return.          
            end.
          end.  
          
          if llgCURunToo 
          then do:
            run spDataLink (
                input linCurrentCompileUnitObjectID,
                input iinTableObjectID,
                input false,
                input true,
                input false,
                input false,
                input true,
                output iinConnectorID
                ).
            if     iinConnectorID eq ?
                or iinConnectorID eq 0
            then do:
              lobExceptionStack:pushException( 564, 2, "CreateConnector failed for = " + ichTableName + " in recordphrase in " + lchCurrentProgramUnitName ).
              return.          
            end.
            run spCreateConnectorTag (
                input iinConnectorID,
                input "WhereClause",
                input ichWhereClause,
                output iinTagID
                ).         
            if     iinTagID eq ?
                or iinTagID eq 0
            then do:
              lobExceptionStack:pushException( 565, 2, "CreateConnectorTag Whereclause failed for = " + ichTableName + " in recordphrase in " + lchCurrentProgramUnitName ).
              return.          
            end.
            run spCreateConnectorTag (
                input iinConnectorID,
                input "LockStatus",
                input ichLockStatus,
                output iinTagID
                ).         
            if     iinTagID eq ?
                or iinTagID eq 0 
            then do:
              lobExceptionStack:pushException( 566, 2, "CreateConnectorTag LockStatus failed for = " + ichTableName + " in recordphrase in " + lchCurrentProgramUnitName ).
              return.          
            end.
          end.  
        end.
/* .......................................................................... */
      when "tableref"                     /* table reference within procedure */
        then do:
          ichTableName = iphnAttributes:get-value-by-qname("table").
          run spFindObject (
              input "Class",
              input "table",
              input ichTableName,
              output iinTableObjectID
              ).
          if     iinTableObjectID eq ?
              or iinTableObjectID eq 0
          then do:
            lobExceptionStack:pushException( 571, 2, "FindObject failed for " + ichTableName + " in tableref in " + lchCurrentProgramUnitName ).
            return.          
          end.
                          /* checking for read/write, but not yet implemented */
                          /* so ilgRead and ilgWrite will always be false     */
          if iphnAttributes:get-value-by-qname("create") eq "true"
          then ilgCreate = true.
          else ilgCreate = false. 
          if iphnAttributes:get-value-by-qname("delete") eq "true"
          then ilgDelete = true.
          else ilgDelete = false. 
          if iphnAttributes:get-value-by-qname("read") eq "true"
          then ilgRead = true.
          else ilgRead = false. 
          if iphnAttributes:get-value-by-qname("write") eq "true"
          then ilgWrite = true.
          else ilgWrite = false. 
          if iphnAttributes:get-value-by-qname("readonly") eq "true"
          then ichStereotype = "oeDataLinkReadOnly".
          else ichStereotype = "oeDataLinkModify".
          
          case lchCurrentParentElementType:
            when "cu"
              then do:
                run spDataLink (
                    linCurrentProgramUnitObjectID,
                    input iinTableObjectID,
                    input ilgCreate,
                    input ilgRead,
                    input ilgWrite,
                    input ilgDelete,
                    input false,
                    output iinConnectorID
                    ).
                if     iinConnectorID eq ?
                    or iinConnectorID eq 0
                then do:
                  lobExceptionStack:pushException( 572, 2, "DataLink failed for " + ichTableName + " in tableref in " + lchCurrentProgramUnitName ).
                  return.          
                end.
              end.
            when "procedure"
              then do:
                run spDataLink (
                    linCurrentProcedureObjectID,
                    input iinTableObjectID,
                    input ilgCreate,
                    input ilgRead,
                    input ilgWrite,
                    input ilgDelete,
                    input false,
                    output iinConnectorID
                    ).
                if     iinConnectorID eq ?
                    or iinConnectorID eq 0
                then do:
                  lobExceptionStack:pushException( 574, 2, "DataLink failed for " + ichTableName + " in tableref in " + lchCurrentProgramUnitName ).
                  return.          
                end.
              end.
            when "include"       /* Provided in the hope of future capability */
              then do:
               lobExceptionStack:pushException( 577, 4, "Tableref found for Include Unit " + lchCurrentIncludeUnitName + " - not yet implemented" ).
              end.
            otherwise 
              do:
                lobExceptionStack:pushException( 578, 2, "Unexpected containing unit type = " + lchCurrentParentElementType + " founc for tableref" ).
              end.
          end case.
          
          if       llgPUDataToo 
              and ( llgTrueDataSummary or lchCurrentParentElementType ne "cu" )
          then do:
            run spDataLink (
                linCurrentProgramUnitObjectID,
                input iinTableObjectID,
                input ilgCreate,
                input ilgRead,
                input ilgWrite,
                input ilgDelete,
                input true,
                output iinConnectorID
                ).
            if     iinConnectorID eq ?
                or iinConnectorID eq 0
            then do:
              lobExceptionStack:pushException( 573, 2, "PU Summary DataLink failed for " + ichTableName + " in tableref in " + lchCurrentProgramUnitName ).
              return.          
            end.
          end.
          
          if llgCUDataToo 
          then do:
            run spDataLink (
                linCurrentCompileUnitObjectID,
                input iinTableObjectID,
                input ilgCreate,
                input ilgRead,
                input ilgWrite,
                input ilgDelete,
                input true,
                output iinConnectorID
                ).
            if     iinConnectorID eq ?
                or iinConnectorID eq 0
            then do:
              lobExceptionStack:pushException( 573, 2, "CU Summary DataLink failed for " + ichTableName + " in tableref in " + lchCurrentProgramUnitName ).
              return.          
            end.
          end.
        end.
/* .......................................................................... */
      when "fieldref"                     /* field reference within procedure */
        then do:
          ichTableName = entry( 1, iphnAttributes:get-value-by-qname("field"), "." ).
          ichFieldName = entry( 2, iphnAttributes:get-value-by-qname("field"), "." ).
          run spFindObject (
              input "Class",
              input "table",
              input ichTableName,
              output iinTableObjectID
              ).
          if     iinTableObjectID eq ?
              or iinTableObjectID eq 0
          then do:
            lobExceptionStack:pushException( 581, 2, "FindObject failed for = " + ichTableName + "." + ichFieldName + " in fieldref in " + lchCurrentProgramUnitName ).
            return.          
          end.
          if iphnAttributes:get-value-by-qname("read") eq "true"
          then ilgRead = true.
          else ilgRead = false. 
          if iphnAttributes:get-value-by-qname("write") eq "true"
          then ilgWrite = true.
          else ilgWrite = false.
          if ilgRead
          then do:
            if ilgWrite
            then ichAccessMode = "read/write".
            else ichAccessMode = "read".
          end.
          else do:
            if ilgWrite
            then ichAccessMode = "write".
            else do:
              ichAccessMode = "Unknown".
              lobExceptionStack:pushException( 582, 2, "fieldref " + ichTableName + "." + ichFieldname + " neither read nor write for " + lchCurrentParentElementType ).
            end.
          end.              

          case lchCurrentParentElementType:
            when "cu"
              then do:
                 run spDataLink (
                    input linCurrentProgramUnitObjectID,
                    input iinTableObjectID,
                    input false,
                    input ilgRead,
                    input ilgWrite,
                    input false,
                    input false,
                    output iinConnectorID
                    ).
                if     iinConnectorID eq ?
                    or iinConnectorID eq 0
                then do:
                  lobExceptionStack:pushException( 583, 2, "DataLink failed for = " + ichTableName + "." + ichFieldName + " in fieldref in " + lchCurrentProgramUnitName ).
                  return.          
                end.
                run spCreateConnectorTag (
                    input iinConnectorID,
                    input "Field",
                    input ichFieldName + " is " + ichAccessMode,
                    output iinTagID
                    ).         
                if     iinTagID eq ?
                    or iinTagID eq 0
                then do:
                  lobExceptionStack:pushException( 584, 2, "CreateConnectorTag failed for = " + ichTableName + "." + ichFieldName + " in fieldref in " + lchCurrentProgramUnitName ).
                  return.          
                end.
              end.
            when "procedure"
              then do:
                run spDataLink (
                    linCurrentProcedureObjectID,
                    input iinTableObjectID,
                    input false,
                    input ilgRead,
                    input ilgWrite,
                    input false,
                    input false,
                    output iinConnectorID
                    ).
                if     iinConnectorID eq ?
                    or iinConnectorID eq 0
                then do:
                  lobExceptionStack:pushException( 587, 2, "DataLink failed for = " + ichTableName + "." + ichFieldName + " in fieldref in " + lchCurrentProgramUnitName ).
                  return.          
                end.
                run spCreateConnectorTag (
                    input iinConnectorID,
                    input "Field",
                    input ichFieldName + " is " + ichAccessMode,
                    output iinTagID
                    ).         
                if     iinTagID eq ?
                    or iinTagID eq 0
                then do:
                  lobExceptionStack:pushException( 588, 2, "CreateConnectorTag Field failed for = " + ichTableName + "." + ichFieldName + " in fieldref in " + lchCurrentProgramUnitName ).
                  return.          
                end.
              end.
            when "include"       /* Provided in the hope of future capability */
              then do:
                                                   /* done by bomaddhandler.p */
              end.
            otherwise 
              do:
                                                   /* done by bomaddhandler.p */
              end.
          end case.          
          
          if       llgPUDataToo 
              and ( llgTrueDataSummary or lchCurrentParentElementType ne "cu" )
          then do:
             run spDataLink (
                linCurrentProgramUnitObjectID,
                input iinTableObjectID,
                input false,
                input ilgRead,
                input ilgWrite,
                input false,
                input true,
                output iinConnectorID
                ).
            if     iinConnectorID eq ?
                or iinConnectorID eq 0
            then do:
              lobExceptionStack:pushException( 585, 2, "PU Summary DataLink failed for = " + ichTableName + "." + ichFieldName + " in fieldref in " + lchCurrentProgramUnitName ).
              return.          
            end.
            run spCreateConnectorTag (
                input iinConnectorID,
                input "Field",
                input ichFieldName + " is " + ichAccessMode,
                output iinTagID
                ).         
            if     iinTagID eq ?
                or iinTagID eq 0
            then do:
              lobExceptionStack:pushException( 586, 2, "CreateConnectorTag Field failed for = " + ichTableName + "." + ichFieldName + " in fieldref in " + lchCurrentProgramUnitName ).
              return.          
            end.
          end.
          
          if llgCUDataToo 
          then do:
             run spDataLink (
                linCurrentCompileUnitObjectID,
                input iinTableObjectID,
                input false,
                input ilgRead,
                input ilgWrite,
                input false,
                input true,
                output iinConnectorID
                ).
            if     iinConnectorID eq ?
                or iinConnectorID eq 0
            then do:
              lobExceptionStack:pushException( 585, 2, "CU Summary DataLink failed for = " + ichTableName + "." + ichFieldName + " in fieldref in " + lchCurrentProgramUnitName ).
              return.          
            end.
            run spCreateConnectorTag (
                input iinConnectorID,
                input "Field",
                input ichFieldName + " is " + ichAccessMode,
                output iinTagID
                ).         
            if     iinTagID eq ?
                or iinTagID eq 0
            then do:
              lobExceptionStack:pushException( 586, 2, "CreateConnectorTag Field failed for = " + ichTableName + "." + ichFieldName + " in fieldref in " + lchCurrentProgramUnitName ).
              return.          
            end.
          end.              
                        
        end.
/* .......................................................................... */
      when "newshared"                      /* new shared object in procedure */
        then do:
        end.
/* .......................................................................... */
      when "sharedref"           /* reference to shared variable in procedure */
        then do:
        end.
/* .......................................................................... */
      when "global"                                 /* global shared variable */
        then do:
        end.
/* .......................................................................... */
      when "localsuper"                 /* this-procedure:add-super-procedure */
        then do:
          ichAdderPU = iphnAttributes:get-value-by-qname("cu").
          ichAdderCU = sfCompileName( ichAdderPU ).

          run spFindObject (
              input "Component",
              input "oeProgram",
              input ichAdderPU,
              output iinAdderObjectID
              ).
          if     iinAdderObjectID eq ?
              or iinAdderObjectID eq 0
          then do:
            lobExceptionStack:pushException( 590, 2, "FindObject failed for = `" + ichAdderPU + "' in adder" ).
            return.          
          end.
          run spCreateConnector (
              input "Dependency",
              input "",
              input "",
              input "",
              input "",
              input iinAdderObjectID,
              input linCurrentProgramUnitObjectID,
              input "oeAddLocalSuper", 
              output iinConnectorID 
              ).
          if     iinConnectorID eq ?
              or iinConnectorID eq 0
          then do:
            lobExceptionStack:pushException( 591, 2, "Create Connector failed for = " + ichAdderPU + " in adder" ).
            return.          
          end.
          
          if      llgPURunToo
              and ( ichAdderProc ne "" or llgTrueRunSummary ) 
          then do:
            if     llgSummaryPUSelfJoins
                or iinAdderObjectID ne linCurrentProgramUnitObjectID
            then do:
              run spCreateConnector (
                  input "Dependency",
                  input "",
                  input "",
                  input "",
                  input "",
                  input iinAdderObjectID,
                  input linCurrentSessionSuperPUID,
                  input "oeAddLocalSuperSummary",
                  output iinConnectorID 
                  ).
              if     iinConnectorID eq ?
                  or iinConnectorID eq 0
              then do:
                lobExceptionStack:pushException( 592, 2, "Create Connector failed for = " + ichAdderPU + " in adder" ).
                return.          
              end.
            end.
          end.  
          
          if llgCURunToo
          then do:
            run spFindObject (
                input "Package",
                input "oeCompileUnit",
                input ichAdderCU,
                output iinAdderObjectID
                ).
            if     iinAdderObjectID eq ?
                or iinAdderObjectID eq 0
            then do:
              lobExceptionStack:pushException( 593, 2, "FindObject failed for = " + ichAdderCU + " in adder" ).
              return.          
            end.
            if     llgSummaryCUSelfJoins
                or iinAdderObjectID ne linCurrentCompileUnitObjectID
            then do:
              run spCreateConnector (
                  input "Dependency",
                  input "",
                  input "",
                  input "",
                  input "",
                  input iinAdderObjectID,
                  input linCurrentCompileUnitObjectID,
                  input "oeAddLocalSuperSummary",
                  output iinConnectorID 
                  ).
              if     iinConnectorID eq ?
                  or iinConnectorID eq 0
              then do:
                lobExceptionStack:pushException( 594, 2, "CreateConnector failed for = " + ichAdderCU + " in adder" ).
                return.          
              end.
            end.
          end.
        end.
/* .......................................................................... */
      otherwise do:
          /* reported by bomaddhandler.p
          lobExceptionStack:pushException( 404, 1, "Unexpected Element Type in BoM = " + ipchElementType ).
          */
        end.
  end case.
end procedure.
/* ************************************************************************** */
procedure EndElement:
    define input  parameter ipchNameSpaceURI as character  no-undo.
    define input  parameter ipchLocalName as character  no-undo.
    define input  parameter ipchElementType as character  no-undo.
  
  if llgDebug then lobExceptionStack:trace( "Start of " + ipchElementType ).  
  case ipchElementType:
      when "bom" then .                                              /* start */
      when "sessionsuper"                           /* Session superprocedure */
        then do:
        end.
      when "adder"                            /* Prog that adds session super */
        then do:
        end.
      when "cu"                                               /* Compile Unit */
        then do:
          lchCurrentProgramUnitName = "".
          lchCurrentCompileUnitName = "".
          lchCurrentParentElementType = "".
        end.
      when "procedure"                            /* Proc within compile unit */
        then do:
          lchCurrentProcedureName = "".
          lchCurrentProcedureType = "".
          lchCurrentParentElementType = "cu".
        end.
      when "include"                   /* reference to include file procedure */
        then do:
        end.
      when "caller"                        /* Program that calls compile unit */
        then do:
        end.
      when "recordphrase"                    /* where clause within procedure */
        then do:
        end.
      when "tableref"                     /* table reference within procedure */
        then do:
        end.
      when "fieldref"                     /* field reference within procedure */
        then do:
        end.
      when "newshared"                    /* new shared variable in procedure */
        then do:
        end.
      when "sharedref"           /* reference to shared variable in procedure */
        then do:
        end.
      when "global"                                 /* global shared variable */
        then do:
        end.
      when "localsuper"                                          /* something */
        then do:
        end.        
      otherwise do:
          /* Reported on by bomaddhandler.p
          lobExceptionStack:pushException( 505, 1, "Unexpected End Element Type in BoM = " + ipchElementType ).
          */
        end.
  end case.
end procedure.
/* ************************************************************************** */         
procedure BoMLinkDestruct:
  delete object lobExceptionStack.
  lobExceptionStack = ?.
end procedure. 
/***************************** End of Program *********************************/