/*------------------------------------------------------------------------------
    File        : codemain.p
    Description : Parse Analyst BoM input and other sources to build UML for ABL code
    Author(s)   : Thomas Mercer-Hursh
    Created     : Fri Jan 04 14:55:41 PST 2008
    Notes       : Assumes that dictionary information is already in the repository
                : Consists of three sections:
                  1. Create list of program names and descriptions from user source (optional)
                  2. Create table of menu structure from user source (optional)
                  3. Read BoM from Analyst
                  The first two will be incoporated into the UML if available
                  Load directory is assumed to be the base directory for where code is run.
                  Change lchLoadDirectory and lchBoM as needed.
  ----------------------------------------------------------------------------*/
using com.cintegrity.Lib.Exception.*.
/* ***************************  Definitions  ******************************** */
                                                                /* Parameters */
define variable lchBoM as character no-undo initial "bom.xml".
                                                                   /* Handles */
define variable lhnBoMAddParser as handle no-undo.
define variable lhnBoMLinkParser as handle no-undo.
define variable lhnBoMAddHandler as handle no-undo.
define variable lhnBoMLinkHandler as handle no-undo.
define variable lhnEAUtilities as handle no-undo.
define variable lchDiskFile as character no-undo.
                                                                   /* Classes */
define variable lobExceptionStack as class SimpleExceptionStack no-undo.
/* *************************  Initialization ******************************** */
                                                         /* Exception Handler */
lobExceptionStack = new SimpleExceptionStack().
                                                           /* Check for Alias */
if not connected( "DictSource" )
then do:
  lobExceptionStack:pushException( 301, 1, "Dictionary Data Source must be connected with alias `DictSource'" ).
  run CleanUp.
  return.
end.

lchDiskFile = search( lchBoM ).
if lchDiskFile eq ?
then do:
  lobExceptionStack:pushException( 302, 1, "Cannot find " + lchBoM + " - aborting" ).
  run CleanUp.
  return.
end.
  
lobExceptionStack:trace( "Running ABL Code to UML routine at " + string( now ) ).
/* ***************************  Main Block  *************************** */
lobExceptionStack:trace("Parse start at " + string(now)).
                                                         /* Read BoM for Adds */
create sax-reader lhnBoMAddParser.
run com/cintegrity/ABLtoUML/code/bomaddhandler.p persistent set lhnBoMAddHandler.
run Initialize in lhnBoMAddHandler.
lhnBoMAddParser:handler = lhnBomAddHandler.
lhnBoMAddParser:set-input-source("file", lchDiskFile ).
lhnBoMAddParser:sax-parse() no-error.
if error-status:error
then do:
  if error-status:num-messages gt 0
  then lobExceptionStack:pushException( 304, 1, error-status:get-message(1)).
  else lobExceptionStack:pushException( 305, 1, return-value ).
end.
delete object lhnBoMAddParser.
run BoMAddDestruct in lhnBoMAddHandler.
delete procedure lhnBoMAddHandler.
assign
    lhnBoMAddParser = ?
    lhnBoMAddHandler = ?
    .
                                                       /* Export log thus far */
lobExceptionStack:writeTextExternal( 9, "ABLtoUMLLog.txt", true ).
lobExceptionStack:clearExceptions().
                                                        /* Read BoM for Links */
create sax-reader lhnBoMLinkParser.
run com/cintegrity/ABLtoUML/code/bomlinkhandler.p persistent set lhnBoMLinkHandler.
run Initialize in lhnBoMLinkHandler.
lhnBoMLinkParser:handler = lhnBomLinkHandler.
lhnBoMLinkParser:set-input-source("file", lchDiskFile ).
lhnBoMLinkParser:sax-parse() no-error.
if error-status:error
then do:
  if error-status:num-messages gt 0
  then lobExceptionStack:pushException( 304, 1, error-status:get-message(1)).
  else lobExceptionStack:pushException( 305, 1, return-value ).
end.
delete object lhnBoMLinkParser.
run BoMLinkDestruct in lhnBoMLinkHandler.
delete procedure lhnBoMLinkHandler.  
assign
    lhnBoMLinkParser = ?
    lhnBoMLinkHandler = ?
    .
                                                       /* Export log thus far */
lobExceptionStack:writeTextExternal( 9, "ABLtoUMLLog.txt", true ).
lobExceptionStack:clearExceptions().   

lobExceptionStack:trace("Parse complete at " + string(now)).
                                           /* Make postprocessing adjustments */
run com/cintegrity/ABLtoUML/code/postprocess.p.
                                                       /* Export log thus far */
lobExceptionStack:writeTextExternal( 9, "ABLtoUMLLog.txt", true ).
lobExceptionStack:clearExceptions().   

run CleanUp.
return.
/* ****************************  Clean Up  ********************************** */
procedure CleanUp:
  delete object lobExceptionStack.
  lobExceptionStack = ?.
end.
/* *************************** End of Program ******************************* */