/*------------------------------------------------------------------------------
    File        : postprocess.p
    Description : Make standard alterations to completed model 
    Author(s)   : Thomas Mercer-Hursh
    Created     : Thu May 01 11:38:25 PDT 2008
    Notes       : Includes: 
                  * Distinguish between public and private oeFunction and oeIP
  ----------------------------------------------------------------------------*/
using com.cintegrity.Lib.Exception.*.
/* ***************************  Definitions  ******************************** */
                                                                   /* Objects */
define variable lobExceptionStack as class SimpleExceptionStack no-undo.
                                                            /* Work variables */
define variable lchTargetContainer as character no-undo.
define variable lchSourceContainer as character no-undo.
define variable llgHasPublic as logical no-undo.
                                                                   /* Buffers */
define buffer lbfObject for Repository.t_Object.
define buffer lbfStartObject for Repository.t_Object.
define buffer lbfConnector for Repository.t_Connector.
/* **************************  Initializations  ***************************** */
lobExceptionStack = new SimpleExceptionStack().
lobExceptionStack:trace( "Start of Private/Public Processing at " + string(now) ).

/* ***************************  Main Block  ********************************* */
                                                /* Private/Public Distinction */
for each lbfObject exclusive-lock
    where lbfObject.Object_Type eq "Component"
    :
  if      lbfObject.Stereotype ne "oeFunctionPrivate"
      and lbfObject.Stereotype ne "oeIPPrivate"
  then next.
  
  assign
      llgHasPublic = false
      lchTargetContainer = entry(1,lbfObject.Name,":")
      .
  for each lbfConnector exclusive-lock
      where lbfConnector.End_Object_ID eq lbfObject.Object_ID
      :
    find lbfStartObject no-lock 
        where lbfStartObject.Object_ID eq lbfConnector.Start_Object_ID   
        .
    if index(lbfStartObject.Name,":") > 0
    then lchSourceContainer = entry(1,lbfStartObject.Name,":").
    else lchSourceContainer = lbfStartObject.Name.
    
    if lchSourceContainer ne lchTargetContainer
    then do:
      lbfConnector.Stereotype = replace(lbfConnector.Stereotype,"Private","Public").
      llgHasPublic = true.
    end.
  end.
  
  if llgHasPublic 
  then lbfObject.Stereotype = replace(lbfObject.Stereotype,"Private","Public").
  
end.

lobExceptionStack:trace( "End of Private/Public Processing at " + string(now) ).

run Cleanup.

/* ************************************************************************** */
procedure CleanUp:
  delete object lobExceptionStack.
  lobExceptionStack = ?.
end.