/*------------------------------------------------------------------------------
    File        : startup.p
    Description : Initial startup procedure for ABL to UML load
    Author(s)   : Thomas Mercer-Hursh
    Created     : Sun Dec 16 12:52:56 PST 2007
    Notes       : Requires:
                  Dictionary database connected with -ld DictSource
                  EA Repository database connected with -ld Repository
                  Dictionary should be connected first, i.e., should be DICTDB
  ----------------------------------------------------------------------------*/
using com.cintegrity.Lib.Exception.*.
/* ***************************  Definitions  ******************************** */
                                                                /* "Paramters */
define variable linLogLevel as integer no-undo initial 9.
                                                                   /* Handles */
define variable lhnCustomSP as handle no-undo. 
define variable lhnXML as handle no-undo.
define variable lhnEAUtilities as handle no-undo.
                                                                   /* Objects */
define variable lobExceptionStack as class SimpleExceptionStack no-undo.
                                                            /* Work Variables */
define variable linEntry as integer no-undo.
define variable llcExceptionLog as longchar no-undo.
define variable llgGotError as logical no-undo initial false.
define variable lchDictSourceName as character no-undo.                           
/* ************************* Initialization ********************************* */
                                                             /* Start New Log */
output to "ABLToUMLLog.txt".
output close.                                                             
                                                         /* Exception Handler */
lobExceptionStack = new SimpleExceptionStack().
lobExceptionStack:trace( "Start of ABL to UML main program at " + string(now) ).  
                                                /* Get Dictionary Source Name */
if not connected( "DictSource" )
then do:
  lobExceptionStack:pushException( 100, 1, "Dictionary Data Source must be connected with alias `DictSource'" ).
  llgGotError = true.
end.
else do:
  lchDictSourceName = pdbname("DictSource").
                                                          /* Check Repository */
  if not connected( "Repository" )
  then do:
    lobExceptionStack:pushException( 101, 1, "EA Repository database must be connected with alias `Repository'" ).
    llgGotError = true.
  end.
  else do:
                                                       /* Start utility super */
    run com/cintegrity/ABLtoUML/util/eautilsp.p persistent set lhnEAUtilities.
    if valid-handle( lhnEAUtilities )
    then do:
      session:add-super-procedure (lhnEAUtilities).
      run spGetInitialValues.
      if lobExceptionStack:numExceptions( 1 ) gt 0
      then do: 
        lobExceptionStack:pushException( 102, 1, "Error in spGetInitialValues - aborting" ).
        llgGotError = true.
      end.
      else do:
        lobExceptionStack:trace( "Databases connected and EA Utility Superprocedure started" ).
                                                            /* Custom code SP */
        run com/cintegrity/ABLtoUML/site/customsp.p persistent set lhnCustomSP.
        if valid-handle( lhnCustomSP )
        then do:
          session:add-super-procedure (lhnCustomSP).
          run spInitializeCustom.
          if lobExceptionStack:numExceptions( 1 ) gt 0
          then do: 
            lobExceptionStack:pushException( 103, 1, "Error in spInitializeCustom - aborting" ).
            llgGotError = true.
          end.
          else do:          
            lobExceptionStack:trace("Initialize site specific utilites and data complete at " + string(now)).
          end.
        end.
        else do:
          lobExceptionStack:pushException( 104, 1, "Failure to Custom Code Utility procedure" ).
          llgGotError = true.      
        end.          
      end.
    end.
    else do:
      lobExceptionStack:pushException( 105, 1, "Failure to start EA Utility procedure" ).
      llgGotError = true.      
    end.  
  end.
end.

/* ***************************  Main Block  *************************** */
                                       /* Build UML from data dictionary data */
if not llgGotError
then do:
  run com/cintegrity/ABLtoUML/dict/dictmain.p.
  if lobExceptionStack:numExceptions( 1 ) gt 0
  then do:
    lobExceptionStack:trace( "Dictionary to UML routine failed - aborting" ).
    llgGotError = true. 
  end.
  else do:
    lobExceptionStack:trace( "Dictionary to UML routine complete at " + string(now) ).
                                           /* Build UML from BoM for ABL Code */
    run com/cintegrity/ABLtoUML/code/codemain.p.
    if lobExceptionStack:numExceptions( 1 ) gt 0
    then do:
      lobExceptionStack:trace( "ABL Code to UML routine failed - aborting" ).
      llgGotError = true.
    end. 
    else do:
      lobExceptionStack:trace( "ABL Code to UML routine complete at " + string(now) ).  
                                                   /* Site Specific Additions */
      run spCustomAdditions.
      if lobExceptionStack:numExceptions( 1 ) gt 0
      then do:
        lobExceptionStack:trace( "Site Specific Additions failed - aborting" ).
        llgGotError = true.
      end. 
      else lobExceptionStack:trace( "Site Specific Additions complete at " + string(now) ).  
    end.
  end.    
end.

if llgGotError
then lobExceptionStack:trace( "ABL to UML main program terminating with error at " + string(now) ). 
else lobExceptionStack:trace( "ABL to UML main program complete at " + string(now) ).  
                                                                /* Finish log */
lobExceptionStack:writeTextExternal( 9, "ABLtoUMLLog.txt", true ).
lobExceptionStack:clearExceptions().
/* ****************************  Clean Up  ********************************** */
if not llgGotError
then do:
  if valid-handle( lhnEAUtilities )
  then do:
    run spEAUtilitiesDestruct.
    session:remove-super-procedure (lhnEAUtilities).
    delete procedure lhnEAUtilities.
    lhnEAUtilities = ?.
  end.
  if valid-handle( lhnCustomSP )
  then do:
    run spCustomDestruct.
    session:remove-super-procedure (lhnCustomSP).
    delete procedure lhnCustomSP.
    lhnCustomSP = ?.
  end.  
end.
delete object lobExceptionStack.
lobExceptionStack = ?.

return.
/* *************************** End of Program ******************************* */