/*------------------------------------------------------------------------------
    File        : eautilsp.p
    Purpose     : Provide some standard utilities for working with an EA repository
    Description : Super procedure for EA related utilities	
    Author(s)   : Thomas Mercer-Hursh
    Created     : Mon Dec 17 11:51:39 PST 2007
    Notes       : Assumes OpenEdge Repository connected with Alias `Repository'
  ----------------------------------------------------------------------------*/
using com.cintegrity.Lib.Exception.*.
/* ***************************  Definitions  ******************************** */
                                                                   /* Top IDs */
define variable linPackageID as integer no-undo initial 1.
define variable lchPackageName as character no-undo initial 1.
define variable linObjectID as integer no-undo initial 1.
define variable linOperationID as integer no-undo initial 1.
define variable linParentPackageID as integer no-undo initial 1.
define variable linAttributeID as integer no-undo initial 1.
define variable linObjectPropertyID as integer no-undo initial 1.
define variable linAttributeTagID as integer no-undo initial 1.
define variable linOperationTagID as integer no-undo initial 1.
define variable linConnectorID as integer no-undo initial 1.
define variable linConnectorTagID as integer no-undo initial 1.
                                                                   /* Objects */
define variable lobExceptionStack as class SimpleExceptionStack no-undo.
                                                               /* Temp-Tables */
define temp-table ttObject
    field chType as character
    field chStereotype as character
    field chName as character
    field inObjectID as integer
    field chEAGUID as character
    index ixByName is primary chType chStereotype chName .
define temp-table ttStereotype
    field chStereotype as character
    field chAppliesTo as character
    field chEAGUID as character
    index ixByStereotype is primary unique chStereotype chAppliesTo.
/* ************************** Function Declarations ************************* */
function sfCompileName returns character ( input ipchProgramName as character ) forward.
function sfGetEAGUID returns character forward:
function sfGetXrefType returns character ( input ipchClientType as character ) forward.
/* ****************************** Main Body ********************************* */

/* ******************************  Methods ********************************** */
procedure spGetInitialValues:
  
  define buffer ibfPackage for Repository.t_Package.
  define buffer ibfObject for Repository.t_Object.
  define buffer ibfAttribute for Repository.t_Attribute.
  define buffer ibfOperation for Repository.t_Operation.
  define buffer ibfOperationParams for Repository.t_OperationParams.
  define buffer ibfObjectProperty for Repository.t_ObjectProperties.
  define buffer ibfAttributeTag for Repository.t_AttributeTag.
  define buffer ibfOperationTag for Repository.t_OperationTag.
  define buffer ibfConnector for Repository.t_Connector.
  define buffer ibfConnectorTag for Repository.t_ConnectorTag.
  define buffer ibfStereotype for Repository.t_Stereotypes.
  
  define variable ichEAGUID as character no-undo.
  define variable iinError as integer no-undo.
                                                          /* Exception object */
  lobExceptionStack = new SimpleExceptionStack().
                                                   /* Check for Repository DB */
  if not connected( "Repository" )
  then do:
    lobExceptionStack:pushException( 901, 1, "Repository Database must be connected with alias `Repository'" ).
    run spEAUtilitiesDestruct.
    return.
  end.
                                              /* Find highest used Package_ID */
  for each ibfPackage no-lock
      by ibfPackage.Package_ID descending
      :
    linPackageID = ibfPackage.Package_ID.
    leave.
  end.
                                               /* Find highest used Object_ID */
  for each ibfObject no-lock
      by ibfObject.Object_ID descending
      :
    linObjectID = ibfObject.Object_ID.
    leave.
  end.
                                            /* Find highest used Operation_ID */
  for each ibfOperation no-lock
      by ibfOperation.OperationID descending
      :
    linOperationID = ibfOperation.OperationID.
    leave.
  end.
                                            /* Find highest used Attribute_ID */
  for each ibfAttribute no-lock
      by ibfAttribute.ID descending
      :
    linAttributeID = ibfAttribute.ID.
    leave.
  end.
                                       /* Find highest used ObjectProperty_ID */
  for each ibfObjectProperty no-lock
      by ibfObjectProperty.PropertyID descending
      :
    linObjectPropertyID = ibfObjectProperty.PropertyID.
    leave.
  end.
                                         /* Find highest used AttributeTag_ID */
  for each ibfAttributeTag no-lock
      by ibfAttributeTag.PropertyID descending
      :
    linAttributeTagID = ibfAttributeTag.PropertyID.
    leave.
  end.
                                         /* Find highest used OperationTag_ID */
  for each ibfOperationTag no-lock
      by ibfOperationTag.PropertyID descending
      :
    linOperationTagID = ibfOperationTag.PropertyID.
    leave.
  end.
                                            /* Find highest used Connector_ID */
  for each ibfConnector no-lock
      by ibfConnector.Connector_ID descending
      :
    linConnectorID = ibfConnector.Connector_ID.
    leave.
  end.
                                         /* Find highest used ConnectorTag_ID */
  for each ibfConnectorTag no-lock
      by ibfConnectorTag.PropertyID descending
      :
    linConnectorTagID = ibfConnectorTag.PropertyID.
    leave.
  end.  
                                              /* Prefill table of stereotypes */
  for each ibfStereotype no-lock:
    create ttStereotype.
    assign 
        ttStereotype.chStereotype = ibfStereotype.Stereotype
        ttStereotype.chAppliesTo  = ibfStereotype.AppliesTo
        ttStereotype.chEAGUID = ibfStereotype.EA_GUID
        no-error.
    if error-status:error
    then do:
      lobExceptionStack:pushException( 902, 2, "Load of Stereotypes failed for " + ibfStereotype.Stereotype + "/" + ibfStereotype.AppliesTo ).
      do iinError = 1 to error-status:num-messages:
        lobExceptionStack:pushException( 903, 4, error-status:get-message( iinError ) ).
      end.     
    end.
  end.
                                 /* Check for internally supplied stereotypes */
  run spFindStereotype (
      input "oeDataLinkReadOnly",
      input "Dependency",
      output ichEAGUID
      ).
  run spFindStereotype (
      input "oeDataLinkModify",
      input "Dependency",
      output ichEAGUID
      ).
  
end procedure.
/* -------------------------------------------------------------------------- */
procedure spFindDataModelPackage:
  define output parameter opinPackageID as integer    no-undo.

  define buffer ibfPackage for Repository.t_Package. 
   
  find ibfPackage no-lock
      where ibfPackage.Name = "Data Model"
      no-error.
  if not available( ibfPackage )
  then do:
    lobExceptionStack:pushException( 904, 1, "Can't find Data Model record in Repository" ).
    return.
  end.
  opinPackageID = ibfPackage.Package_ID.
end procedure.  
/* -------------------------------------------------------------------------- */
procedure spFindComponentModelPackage:
  define output parameter opinPackageID as integer    no-undo.

  define buffer ibfPackage for Repository.t_Package. 
   
  find ibfPackage no-lock
      where ibfPackage.Name = "Component Model"
      no-error.
  if not available( ibfPackage )
  then do:
    lobExceptionStack:pushException( 905, 1, "Can't find Component Model record in Repository" ).
    return.
  end.
  opinPackageID = ibfPackage.Package_ID.
end procedure.  
/* -------------------------------------------------------------------------- */
procedure spFindUIModelPackage:
  define output parameter opinPackageID as integer    no-undo.

  define buffer ibfPackage for Repository.t_Package. 
   
  find ibfPackage no-lock
      where ibfPackage.Name = "User Interface Model"
      no-error.
  if not available( ibfPackage )
  then do:
    lobExceptionStack:pushException( 906, 1, "Can't find User Interface Model record in Repository" ).
    return.
  end.
  opinPackageID = ibfPackage.Package_ID.
end procedure.  
/* -------------------------------------------------------------------------- */
procedure spCreatePackage:
  define input  parameter ipchPackageName as character  no-undo.
  define input  parameter ipinParentPackageID as integer    no-undo.
  define output parameter opinPackageID as integer    no-undo.
  
  define buffer ibfPackage for Repository.t_Package. 
  
  for each ibfPackage no-lock
      where ibfPackage.Parent_ID eq ipinParentPackageID
      :
    if ibfPackage.Name eq ipchPackageName
    then do:
      lobExceptionStack:pushException( 911, 2, "Attempt to add package " + ipchPackageName + " with parent " + string(ipinParentPackageID) + ", but a package of that name and parent already exists" ).
      opinPackageID = ?.
      return.
    end.
  end.
  
  if ipinParentPackageID eq ?
  then do:
    lobExceptionStack:pushException( 912, 1, "Attempt to create package " + ipchPackageName + " with unknown parent ID" ).
    return.
  end.
    
  create ibfPackage.
  assign
      linPackageID = linPackageID + 1
      opinPackageID = linPackageID
      ibfPackage.EA_GUID = sfGetEAGUID()
      ibfPackage.Package_ID = linPackageID
      ibfPackage.Name = ipchPackageName
      ibfPackage.Parent_ID = ipinParentPackageID
      ibfPackage.CreatedDate = now     
      ibfPackage.ModifiedDate = now   
      /* ibfPackage.Notes = */
      /* XibfPackage.MLPath = "" */
      ibfPackage.IsControlled = 0
      ibfPackage.LastLoadDate = now   
      ibfPackage.LastSaveDate = now   
      ibfPackage.Version = "1.0"
      /* ibfPackage.Protected = 0 */
      /* ibfPackage.PkgOwner = "" */
      /* ibfPackage.UMLVersion = "2.1" */
      /* ibfPackage.UseDTD = 0 */
      /* ibfPackage.LogXML = 0 */
      /* ibfPackage.CodePath = "" */
      /* ibfPackage.NameSpace = "" */
      /* ibfPackage.Tpos = n */
      /* ibfPackage.PackageFlags = "" */
      /* ibfPackage.BatchSave = 0 */
      /* ibfPackage.BatchLoad = 0 */
      .
end procedure.        
/* -------------------------------------------------------------------------- */
procedure spAddPackageNotes:
  define input  parameter ipinPackageID as integer no-undo.
  define input  parameter ipchNotes as character  no-undo.
 
  define buffer ibfPackage for Repository.t_Package.  
  
  if ipinPackageID eq ?
  then do:
    lobExceptionStack:pushException( 913, 2, "Attempt to add notes to unknown package ID" ).
    if ipchNotes ne ?
    then do:
      lobExceptionStack:pushException( 914, 4, "Notes = " + ipchNotes ).    
    end.
    return.
  end.
    
  find ibfPackage exclusive-lock
      where ibfPackage.Package_ID = ipinPackageID
      no-error.
  if available( ibfPackage )
  then do:
    ibfPackage.Notes = ipchNotes.
  end.
  else do:
    lobExceptionStack:pushException( 915, 2, "Attempt to add Notes to non-existent Package = " + string(ipinPackageID) ).
    lobExceptionStack:pushException( 916, 4, "Notes = " + ipchNotes ).
  end.
end procedure.
/* -------------------------------------------------------------------------- */
procedure spCreateObject:
  define input  parameter ipchObjectType as character  no-undo.
  define input  parameter ipchObjectName as character  no-undo.
  define input  parameter ipinPackageID as integer    no-undo.
  define input  parameter ipinParentID as integer    no-undo.
  define input  parameter ipchStereotype as character  no-undo.
  define input  parameter ipchNotes as character  no-undo.
  define input parameter ipchAlias as character no-undo.
  define output parameter opinObjectID as integer    no-undo.
  
  define buffer ibfObject for Repository.t_Object.

  define variable ichEAGUID as character no-undo.
  define variable ihnBuffer as handle no-undo.
  define variable ihnField as handle no-undo.

  if ipchAlias eq ?
  then do:
    lobExceptionStack:pushException( 921, 2, "Attempt to add unknown Alias to object " + ipchObjectName + " in spCreateObject" ).
    ipchAlias = "".
  end.
  if ipchNotes eq ?
  then do:
    lobExceptionStack:pushException( 922, 2, "Attempt to add unknown Notes to object " + ipchObjectName + " in spCreateObject" ).
    ipchNotes = "".
  end.
  
  find ttObject
      where ttObject.chType eq ipchObjectType
      and   ttObject.chStereotype eq ipchStereotype
      and   ttObject.chName eq ipchObjectName
      no-error.
  if not available( ttObject )
  then do:
    create ibfObject.
    assign
        linObjectID = linObjectID + 1
        opinObjectID = linObjectID
        ibfObject.EA_GUID = sfGetEAGUID()
        ibfObject.Object_ID = linObjectID
        ibfObject.Object_Type = ipchObjectType
        /* ibfObject.Diagram_ID */
        ibfObject.Name = ipchObjectName
        /* ibfObject.Alias = ipchAlias see below */
        ibfObject.Author = "com.cintegrity.ADBtoUML.dict.eautilsp.p"
        ibfObject.Version = "1.0"
        ibfObject.Note = ""
        ibfObject.Package_ID = ipinPackageID
        ibfObject.Stereotype = ipchStereotype
        ibfObject.Ntype = 0
        /* ibfObject.Effort */
        /* ibfObject.Style */
        ibfObject.Backcolor = -1
        ibfObject.BorderStyle = 0
        ibfObject.Borderwidth = -1
        ibfObject.FontColor = -1
        ibfObject.Bordercolor = -1         
        ibfObject.CreatedDate = now         
        ibfObject.ModifiedDate = now        
        ibfObject.Status = "Implemented"          
        /* ibfObject.Abstract = 0 */          
        /* ibfObject.Tagged */
        ibfObject.Pdata1 = string(ipinPackageID) when (ipchObjectType = "Package" or ipchObjectType = "Text")
        ibfObject.Pdata2 = "OpenEdge" when ipchStereoType = "table"
        ibfObject.Pdata3 = "0" when ipchObjectType = "Text"
        ibfObject.Pdata4 = "0" when ipchStereoType = "table"
        /* ibfObject.Pdata5 */
        /* ibfObject.Concurrency */
        /* ibfObject.Visibility */
        /* ibfObject.Persistence */
        /* ibfObject.Cardinality */
        ibfObject.Gentype = "OpenEdge ABL"
        /* ibfObject.Genfile */
        /* ibfObject.Header1 */
        /* ibfObject.Header2 */
        ibfObject.Phase = "1.0"
        ibfObject.Scope = "Public"
        /* ibfObject.Genoption */
        /* ibfObject.Genlinks */
        /* ibfObject.Classifier */
        ibfObject.ParentID = ipinParentID
        /* ibfObject.Runstate */
        /* ibfObject.Classifier_GUID */
        /* ibfObject.Tpos */
        /* ibfObject.IsRoot */
        /* ibfObject.IsLeaf */
        /* ibfObject.IsActive */
        /* ibfObject.StateFlags */
        /* ibfObject.Multiplicity */
        /* ibfObject.StyleEx */
        /* ibfObject.ActionFlags */
        /* ibfObject.EventFlags */
        ihnBuffer = buffer ibfObject:handle
        ihnfield = ihnbuffer:BUFFER-FIELD(5)
        ihnField:BUFFER-VALUE = ipchAlias
        .
    create ttObject.
    assign
        ttObject.chType = ipchObjectType
        ttObject.chStereotype = ipchStereotype
        ttObject.chName = ipchObjectName
        ttObject.inObjectID = linObjectID
        ttObject.chEAGUID = ibfObject.EA_GUID
        .
  end.       
  if ipchStereotype ne ""
  then do:
    run spFindStereotype (
        input ipchStereotype,
        input ipchObjectType,
        output ichEAGUID
        ).
    run spMakeStereotypeXref (
        input ipchObjectType,
        input ipchStereotype,
        input ttObject.chEAGUID
        ).
  end.
  opinObjectID = ttObject.inObjectID.
end procedure.      
/* -------------------------------------------------------------------------- */
procedure spFindObject:
  define input  parameter ipchObjectType as character no-undo.
  define input parameter ipchStereotype as character no-undo.
  define input  parameter ipchObjectName as character no-undo.
  define output parameter opinObjectID as integer no-undo.
  
  find ttObject
      where ttObject.chType eq ipchObjectType
      and   ttObject.chStereotype eq ipchStereotype
      and   ttObject.chName eq ipchObjectName
      no-error.
  if ambiguous ( ttObject )
  then do:
    lobExceptionStack:pushException( 924, 2, "More than one object with type = " + ipchObjectType + ", stereotype = " + ipchStereotype + " and name = " + ipchObjectName ).
    opinObjectID = ?.
    return.
  end.
  if not available ( ttObject )
  then do:
    /* Handle in caller
    lobExceptionStack:pushException( 909, 2, "No object with type = " + ipchObjectType + ", stereotype = " + ipchStereotype + " and name = " + ipchObjectName + " found" ).
    */
    opinObjectID = ?.
    return.
  end.
  opinObjectID = ttObject.inObjectID.
end procedure.
/* -------------------------------------------------------------------------- */
procedure spAddObjectNotes:
  define input  parameter ipinObjectID as integer no-undo.
  define input  parameter ipchNotes as character  no-undo.
 
  define buffer ibfObject for Repository.t_Object.  
  
  if ipchNotes eq ?
  then do:
    lobExceptionStack:pushException( 925, 2, "Attempt to add unknown Notes to object " + string(ipinObjectID) ).
    ipchNotes = "".
  end.
  
  find ibfObject exclusive-lock
      where ibfObject.Object_ID = ipinObjectID
      no-error.
  if available( ibfObject )
  then do:
    ibfObject.Note = ipchNotes.
  end.
  else do:
    lobExceptionStack:pushException( 926, 2, "Attempt to add Notes to non-existent Object = " + string(ipinObjectID) ).
    lobExceptionStack:pushException( 927, 4, "Notes = " + ipchNotes ).
  end.
end procedure.
/* -------------------------------------------------------------------------- */
procedure spAddObjectAlias:
  define input  parameter ipinObjectID as integer no-undo.
  define input  parameter ipchAlias as character  no-undo.
 
  define buffer ibfObject for Repository.t_Object.  
  define variable ihnBuffer as handle no-undo.
  define variable ihnField as handle no-undo.
  
  if ipchAlias eq ?
  then do:
    lobExceptionStack:pushException( 928, 2, "Attempt to add unknown Alias to object " + string(ipinObjectID) ).
    ipchAlias = "".
  end.
  
  find ibfObject exclusive-lock
      where ibfObject.Object_ID = ipinObjectID
      no-error.
  if available( ibfObject )
  then do:
    assign
        ihnBuffer = buffer ibfObject:handle
        ihnfield = ihnbuffer:BUFFER-FIELD(5)
        ihnField:BUFFER-VALUE = ipchAlias
        .
  end.
  else do:
    lobExceptionStack:pushException( 929, 2, substitute("Attempt to add Alias `$1' to non-existent Object = $2", ipchAlias, string(ipinObjectID) ) ).
  end.
end procedure.
/* -------------------------------------------------------------------------- */
procedure spCreateAttribute:
  define input parameter ipinObjectID as integer no-undo.
  define input parameter ipchObjectName as character no-undo.
  define input parameter ipchStereotype as character no-undo.
  define input parameter ipchNotes as character no-undo.
  define input parameter ipinPos as integer no-undo.
  define input parameter ipinNotNull as integer no-undo.
  define input parameter ipchDefault as character no-undo.
  define input parameter ipchDataType as character no-undo.
  define input parameter ipchAttributeType as character no-undo.
  define output parameter opinAttributeID as integer no-undo.

  define buffer ibfAttribute for Repository.t_Attribute.

  define variable ichEAGUID as character no-undo.

  create ibfAttribute.
  assign
      linAttributeID = linAttributeID + 1
      opinAttributeID = linAttributeID
      ibfAttribute.EA_GUID = sfGetEAGUID()
      ibfAttribute.Object_ID = ipinObjectID
      ibfAttribute.Name = ipchObjectName
      ibfAttribute.Scope = "Public"
      ibfAttribute.Stereotype = ipchStereotype
      /* ibfAttribute.Containment */
      /* ibfAttribute.IsStatic =                  IsUnique    How to set */
      /* ibfAttribute.IsCollection            Source of FK    How to set */
      /* ibfAttribute.IsOrdered               Is PK           How to set */ 
      ibfAttribute.AllowDuplicates = ipinNotNull
      /* ibfAttribute.LowerBound */
      /* ibfAttribute.UpperBound */
      /* ibfAttribute.Container */
      ibfAttribute.Notes = ipchNotes
      /* ibfAttribute.Derived */
      ibfAttribute.ID = linAttributeID
      ibfAttribute.Pos = ipinPos
      /* ibfAttribute.GenOption */
      /* ibfAttribute.Length */
      /* ibfAttribute.Precision */
      /* ibfAttribute.Scale */
      /* ibfAttribute.Const */
      /* ibfAttribute.Classifier */
      ibfAttribute.Default = ipchDefault
      ibfAttribute.Type = ipchDataType
      /* ibfAttribute.StyleEx */
      .
  if ipchStereotype ne ""
  then do:
    run spFindStereotype (
        input ipchStereotype,
        input ipchAttributeType,
        output ichEAGUID
        ).
    run spMakeStereotypeXref (
        input ipchAttributeType,
        input ipchStereotype,
        input ibfAttribute.EA_GUID
        ).
  end.
end procedure.
/* -------------------------------------------------------------------------- */
procedure spCreateOperation:
  define input  parameter ipinObjectID as integer    no-undo.
  define input  parameter ipchOperationName as character  no-undo.
  define input  parameter ipchStereotype as character  no-undo.
  define output parameter opinOperationID as integer    no-undo.
  
  define buffer ibfOperation for Repository.t_Operation.

  define variable ichEAGUID as character no-undo.
  define variable ichOperationType as character no-undo.
    
  create ibfOperation.
  assign
      linOperationID = linOperationID + 1
      opinOperationID = linOperationID
      ibfOperation.EA_GUID = sfGetEAGUID()
      ibfOperation.OperationID = linOperationID
      ibfOperation.Object_ID = ipinObjectID
      ibfOperation.Name = ipchOperationName
      ibfOperation.Scope = "Public"
      /* ibfOperation.Type */
      /* ibfOperation.ReturnArray */
      ibfOperation.Stereotype = ipchStereotype
      /* ibfOperation.IsStatic */
      /* ibfOperation.Concurrency */
      /* ibfOperation.Notes */
      /* ibfOperation.Behavior */
      /* ibfOperation.Abstract */
      /* ibfOperation.GenOption */
      /* ibfOperation.Synchronized */
      /* ibfOperation.Pos */
      /* ibfOperation.Const */
      /* ibfOperation.Style */
      /* ibfOperation.Pure */
      /* ibfOperation.Throws */
      /* ibfOperation.Classifier */
      /* ibfOperation.Code */
      /* ibfOperation.IsRoot */
      /* ibfOperation.IsLeaf */
      /* ibfOperation.IsQuery */
      /* ibfOperation.StateFlags */
      /* ibfOperation.StyleEx */
      .
  if ipchStereotype eq "PK"
      or ipchStereotype eq "FK"
      or ipchStereotype eq "index"
  then ichOperationType = "OpTable".
  else ichOperationType = "Operation".
  
  if ipchStereotype ne ""
  then do:
    run spFindStereotype (
        input ipchStereotype,
        input ichOperationType,
        output ichEAGUID
        ).
    run spMakeStereotypeXref (
        input ichOperationType,
        input ipchStereotype,
        input ibfOperation.EA_GUID
        ).
  end.
end procedure.  
/* -------------------------------------------------------------------------- */
procedure spCreateOperationParam:
  define input  parameter ipinOperationID as integer    no-undo.
  define input  parameter ipchName as character  no-undo.
  define input  parameter ipchType as character  no-undo.
  define input  parameter ipchDefault as character  no-undo.
  define input  parameter ipinPos as integer    no-undo.
  define input  parameter ipchKind as character  no-undo.
  define input  parameter ipchStereotype as character  no-undo.
  
  define buffer ibfOperationParam for Repository.t_OperationParams.

  define variable ichEAGUID as character no-undo.
 
  create ibfOperationParam.
  assign
      ibfOperationParam.EA_GUID = sfGetEAGUID()
      ibfOperationParam.OperationID = ipinOperationID
      ibfOperationParam.Name = ipchName 
      ibfOperationParam.Type = ipchType
      ibfOperationParam.Default = ipchDefault
      ibfOperationParam.Notes = "<<" + ipchStereotype + ">>"  
      ibfOperationParam.Pos = ipinPos
      /* ibfOperationParam.Const */
      /* ibfOperationParam.Style */
      ibfOperationParam.Kind = ipchKind
      /* ibfOperationParam.Classifier */
      ibfOperationParam.StyleEx = ""
      .
  /* Not a true stereotype
  if ipchStereotype ne ""
  then do:
    run spFindStereotype (
        input ipchStereotype,
        input ipchOperationParamType,
        output ichEAGUID
        ).
  end.
  */
end procedure.
/* -------------------------------------------------------------------------- */
procedure spCreateOperationPost:
  define input  parameter ipinOperationID as integer    no-undo.
  define input  parameter ipchCondition as character  no-undo.
  define input  parameter ipchType as character  no-undo.
  
  define buffer ibfOperationPost for Repository.t_OperationPosts.
  
  create ibfOperationPost.
  assign
      ibfOperationPost.OperationID = ipinOperationID
      ibfOperationPost.PostCondition = ipchCondition
      ibfOperationPost.Type = ipchType
      /* ibfOperationPost.Notes */
      .
end procedure.
/* -------------------------------------------------------------------------- */
procedure spCreateObjectTag:
  define input  parameter ipinObjectID as integer    no-undo.
  define input  parameter ipchProperty as character  no-undo.
  define input  parameter ipchValue as character  no-undo.
  define output parameter opinObjectPropertyID as integer    no-undo.
  
  define buffer ibfObjectProperty for Repository.t_ObjectProperties.
  define variable ihnBuffer as handle     no-undo.
  define variable ihnField as handle no-undo.
  define variable ichNote as character no-undo.

  if length( ipchValue ) gt 255
  then do:
    ichNote = ipchValue.
    ipchValue = "<memo>".
  end.
  
  find ibfObjectProperty exclusive-lock
      where ibfObjectProperty.Object_ID = ipinObjectID
      and   ibfObjectProperty.Property = ipchProperty
      no-error.
  if not available ( ibfObjectProperty )
  then do:
    create ibfObjectProperty.  
    assign
        linObjectPropertyID  = linObjectPropertyID + 1
        ibfObjectProperty.EA_GUID = sfGetEAGUID()
        opinObjectPropertyID = linObjectPropertyID
        ibfObjectProperty.PropertyID = linObjectPropertyID
        ibfObjectProperty.Object_ID = ipinObjectID
        ibfObjectProperty.Property = ipchProperty
        .  
  end.
  assign
                                                   /* ibfObjectProperty.Value */
      ihnBuffer = buffer ibfObjectProperty:handle
      ihnfield = ihnbuffer:BUFFER-FIELD(4)
      ihnField:BUFFER-VALUE = ipchValue
      ibfObjectProperty.Notes = ichNote
      .
end procedure.       
/* -------------------------------------------------------------------------- */
procedure spCreateColumnTag:
  define input  parameter ipinAttributeID as integer    no-undo.
  define input  parameter ipchProperty as character  no-undo.
  define input  parameter ipchValue as character  no-undo.
  define output parameter opinAttributeTagID as integer    no-undo.
  
  define buffer ibfAttributeTag for Repository.t_AttributeTag.
  define variable ihnBuffer as handle     no-undo.
  define variable ihnField as handle no-undo.
  define variable ichNote as character no-undo.

  if length( ipchValue ) gt 255
  then do:
    ichNote = ipchValue.
    ipchValue = "<memo>".
  end.
         
  create ibfAttributeTag.
  assign
      linAttributeTagID  = linAttributeTagID + 1
      opinAttributeTagID = linAttributeTagID
      ibfAttributeTag.EA_GUID = sfGetEAGUID()
      ibfAttributeTag.PropertyID = linAttributeTagID
      ibfAttributeTag.ElementID = ipinAttributeID
      ibfAttributeTag.Property = ipchProperty
      ihnBuffer = buffer ibfAttributeTag:handle
      ihnfield = ihnbuffer:BUFFER-FIELD(4)
      ihnField:BUFFER-VALUE = ipchValue
      ibfAttributeTag.Notes = ichNote
      .
end procedure.       
/* -------------------------------------------------------------------------- */
procedure spCreateOperationTag:
  define input  parameter ipinOperationID as integer    no-undo.
  define input  parameter ipchProperty as character  no-undo.
  define input  parameter ipchValue as character  no-undo.
  define output parameter opinOperationTagID as integer    no-undo.
  
  define buffer ibfOperationTag for Repository.t_OperationTag.
  define variable ihnBuffer as handle     no-undo.
  define variable ihnField as handle no-undo.
  define variable ichNote as character no-undo.

  if length( ipchValue ) gt 255
  then do:
    ichNote = ipchValue.
    ipchValue = "<memo>".
  end.
       
  create ibfOperationTag.
  assign
      linOperationTagID  = linOperationTagID + 1
      opinOperationTagID = linOperationTagID
      ibfOperationTag.EA_GUID = sfGetEAGUID()
      ibfOperationTag.PropertyID = linOperationTagID
      ibfOperationTag.ElementID = ipinOperationID
      ibfOperationTag.Property = ipchProperty
      ihnBuffer = buffer ibfOperationTag:handle
      ihnfield = ihnbuffer:BUFFER-FIELD(4)
      ihnField:BUFFER-VALUE = ipchValue
      ibfOperationTag.Notes = ichNote
      .
end procedure.
/* -------------------------------------------------------------------------- */
procedure spCreateConnector:
  define input  parameter ipchConnectorType as character  no-undo.
  define input  parameter ipchSourceCardinality as character  no-undo.
  define input  parameter ipchDestinationCardinality as character  no-undo.
  define input  parameter ipchSourceRole as character  no-undo.
  define input  parameter ipchDestinationRole as character  no-undo.
  define input  parameter ipinStartObjectID as integer    no-undo.
  define input  parameter ipinEndObjectID as integer    no-undo.
  define input  parameter ipchStereotype as character  no-undo.
  define output parameter opinConnectorID as integer    no-undo.
  
  define buffer ibfConnector for Repository.t_Connector.

  define variable ichEAGUID as character no-undo.
  define variable iinError as integer no-undo.
  
  if     ipinStartObjectID eq ?
      or ipinStartObjectID eq 0
      or ipinEndObjectID eq ?
      or ipinEndObjectID eq 0
  then do:
    lobExceptionStack:pushException( 941, 2, substitute("Cannot make &1 connection from object &2 to object &3 (unknown endpoint)", ipchConnectorType, ipinStartObjectID, ipinEndObjectID) ).
    opinConnectorID = ?.
    return.
  end. 
  
  for each ibfConnector exclusive-lock
      where ibfConnector.Start_Object_ID eq ipinStartObjectID
      and   ibfConnector.End_Object_ID eq ipinEndObjectID
      and   ibfConnector.Connector_Type eq ipchConnectorType
      :
    if ibfConnector.Stereotype eq ipchStereotype
    then do:
      assign
          opinConnectorID = ibfConnector.Connector_ID
          ibfConnector.SourceCard = ipchSourceCardinality
          ibfConnector.DestCard = ipchDestinationCardinality
          ibfConnector.SourceRole = ipchSourceRole            /* For FK Links */
          ibfConnector.DestRole = ipchDestinationRole               /* For PK */
          .
      return.
    end.
  end.    

  create ibfConnector.
  assign
      linConnectorID = linConnectorID + 1
      ibfConnector.EA_GUID = sfGetEAGUID()
      ibfConnector.Connector_ID = linConnectorID
      /* ibfConnector.Name */
      ibfConnector.Direction = "Source -> Destination"
      /* ibfConnector.Notes */
      ibfConnector.Connector_Type = ipchConnectorType
      /* ibfConnector.SubType */
      ibfConnector.SourceCard = ipchSourceCardinality
      ibfConnector.SourceAccess = "Public"
      /* ibfConnector.SourceElement */
      ibfConnector.DestCard = ipchDestinationCardinality
      ibfConnector.DestAccess = "Public"
      /* ibfConnector.DestElement */
      ibfConnector.SourceRole = ipchSourceRole                /* For FK Links */
      /* ibfConnector.SourceRoleType */
      /* ibfConnector.SourceRoleNote */
      ibfConnector.SourceContainment = "Unspecified"
      ibfConnector.SourceIsAggregate = 0
      ibfConnector.SourceIsOrdered = 0
      /* ibfConnector.SourceQualifier */
      ibfConnector.DestRole = ipchDestinationRole                   /* For PK */
      /* ibfConnector.DestRoleType */
      /* ibfConnector.DestRoleNote */
      ibfConnector.DestContainment = "Unspecified"
      ibfConnector.DestIsAggregate = 0
      ibfConnector.DestIsOrdered = 0
      /* ibfConnector.DestQualifier */
      ibfConnector.Start_Object_ID = ipinStartObjectID
      ibfConnector.End_Object_ID = ipinEndObjectID
      /* ibfConnector.Top_Start_Label */
      /* ibfConnector.Top_Mid_Label */
      /* ibfConnector.Top_End_Label */
      /* ibfConnector.Btm_Start_Label */
      /* ibfConnector.Btm_Mid_Label */
      /* ibfConnector.Btm_End_Label */
      /* ibfConnector.Start_Edge */
      /* ibfConnector.End_Edge */
      /* ibfConnector.PTStartX */
      /* ibfConnector.PTStartY */
      /* ibfConnector.PTEndX */
      /* ibfConnector.PEEdnY */
      /* ibfConnector.SeqNo */
      /* ibfConnector.HeadStyle */
      /* ibfConnector.LineStyle */
      /* ibfConnector.RouteStyle */
      /* ibfConnector.IsBold */
      /* ibfConnector.LineColor */
      ibfConnector.StereoType = ipchStereotype
      /* ibfConnector.VirtualInheritance */
      /* ibfConnector.LinkAccess */
      /* ibfConnector.Pdata1 */
      /* ibfConnector.Pdata2 */
      /* ibfConnector.Pdata3 */
      /* ibfConnector.Pdata4 */
      /* ibfConnector.Pdata5 */    /* Applies to Data Table Links */
      /* ibfConnector.DiagramID */
      /* ibfConnector.SourceConstraint */
      /* ibfConnector.DestConstraint */
      /* ibfConnector.SourceIsNavigable */
      /* ibfConnector.DestIsNavigable */
      /* ibfConnector.IsRoot */
      /* ibfConnector.IsLeaf */
      /* ibfConnector.IsSpec */
      ibfConnector.SourceChangeable = "none"
      ibfConnector.DestChangeable = "none"
      ibfConnector.SourceTS = "instance"
      ibfConnector.DestTS = "instance"
      /* ibfConnector.StateFlags */
      /* ibfConnector.ActionFlags */
      /* ibfConnector.IsSignal */
      /* ibfConnector.IsStimulus */
      /* ibfConnector.DispatchAction */
      /* ibfConnector.Target2 */
      /* ibfConnector.StyleEx */ /* May apply for FK links */ 
      /* ibfConnector.SourceStereotype */
      /* ibfConnector.DestStereotype */
      ibfConnector.SourceStyle = "Union=0;Derived=0;AllowDuplicates=0;Owned=0;Navigable=Non-Navigable;" 
      ibfConnector.DestStyle = "Union=0;Derived=0;AllowDuplicates=0;Owned=0;Navigable=Navigable;"
      /* ibfConnector.EventFlags */
      no-error.
  if error-status:error
  then do:
    lobExceptionStack:pushException( 942, 2, substitute("Creation of Connector failed for &1 connection from object &2 to object &3", ipchConnectorType, ipinStartObjectID, ipinEndObjectID) ).
    do iinError = 1 to error-status:num-messages:
      lobExceptionStack:pushException( 943, 4, error-status:get-message( iinError ) ).
    end.     
    return.
  end.
  assign
      opinConnectorID = ibfConnector.Connector_ID
      .
  if ipchStereotype ne ""
  then do:
    run spFindStereotype (
        input ipchStereotype,
        input ipchConnectorType,
        output ichEAGUID
        ).
    run spMakeStereotypeXref (
        input ipchConnectorType,
        input ipchStereotype,
        input ibfConnector.EA_GUID
        ).
  end.
  
end procedure.  
/* -------------------------------------------------------------------------- */
procedure spDataLink:
  define input parameter ipinStartObjectID as integer    no-undo.
  define input parameter ipinEndObjectID as integer    no-undo.
  define input parameter iplgHasCreate as logical    no-undo.
  define input parameter iplgHasRead as logical    no-undo.
  define input parameter iplgHasUpdate as logical    no-undo.
  define input parameter iplgHasDelete as logical    no-undo.
  define input parameter iplgIsSummary as logical no-undo.
  define output parameter opinConnectorID as integer    no-undo.  

  define buffer ibfConnector for Repository.t_Connector.
  define buffer ibfConnectorTag for Repository.t_Connectortag.
  
  define variable iinConnectorID as integer    no-undo.
  define variable iinConnectorTagID as integer    no-undo.
  define variable ichFileMode as character  no-undo.
  define variable ihnBuffer as handle     no-undo.
  define variable ihnField as handle no-undo.
  define variable ichStereotype as character no-undo.
  define variable ichNewStereotype as character no-undo.

  for each ibfConnector exclusive-lock
      where ibfConnector.Start_Object_ID eq ipinStartObjectID
      and   ibfConnector.End_Object_ID eq ipinEndObjectID
      and   ibfConnector.Connector_Type eq "Dependency"
      :
    if ibfConnector.Stereotype begins "oeDataLink"
    then do: 
      if     ( iplgIsSummary and ibfConnector.Stereotype matches "*Summary" )
          or ( not iplgIsSummary and not ibfConnector.Stereotype matches "*Summary" )
      then leave.
    end.
  end.     
  
  if iplgIsSummary 
  then ichStereotype = "oeDataLinkReadOnlySummary".
  else ichStereotype = "oeDataLinkReadOnly". 

  if not available ( ibfConnector )
  then do:
    run spCreateConnector (
        input "Dependency",
        input "",
        input "",
        input "",
        input "",
        input ipinStartObjectID,
        input ipinEndObjectID,
        input ichStereotype,
        output iinConnectorID 
        ).
    find ibfConnector exclusive-lock
        where ibfConnector.Connector_ID = iinConnectorID
        no-error.   
    if     not available( ibfConnector )
        or iinConnectorID eq ?
        or iinConnectorID eq 0    
    then do:
      lobExceptionStack:pushException( 944, 2, substitute("Creation of Connector failed for oeDataLinkReadOnly connection from object &1 to object &2", ipinStartObjectID, ipinEndObjectID) ).
      return.
    end.
  end.
  
  if  ibfConnector.Stereotype begins "oeDataLinkReadOnly"
      and ( iplgHasUpdate or iplgHasCreate or iplgHasDelete )
  then do:
    if iplgIsSummary 
    then ichNewStereotype = "oeDataLinkModifySummary".
    else ichNewStereotype = "oeDataLinkModify".     
    run spChangeStereotypeXref (
        input "Dependency", 
        input ibfConnector.Stereotype,
        input ichNewStereotype,
        input iinConnectorID
        ).
   ibfConnector.Stereotype = ichNewStereotype.
  end.
  
  opinConnectorID = ibfConnector.Connector_ID.
  
  find ibfConnectorTag exclusive-lock
      where ibfConnectorTag.ElementID eq opinConnectorID
      and   ibfConnectorTag.Property eq "oeFileMode"
      no-error.
  if not available ( ibfConnectorTag )
  then do:
    run spCreateConnectorTag (
        input opinConnectorID,
        input "oeFileMode",
        input "____",
        output iinConnectorTagID
        ).
    find ibfConnectorTag exclusive-lock
        where ibfConnectorTag.PropertyID eq iinConnectorTagID
        no-error.   
    if     not available( ibfConnectorTag )
        or iinConnectorTagID eq ?
        or iinConnectorTagID eq 0
    then do:
      lobExceptionStack:pushException( 945, 2, substitute("Creation of oeFileMode Connector tag failed for oeDataLinkReadOnly connection from object &1 to object &2", ipinStartObjectID, ipinEndObjectID) ).
      return.                                       
    end.
  end.
  assign
      ihnBuffer = buffer ibfConnectorTag:handle
      ihnField = ihnbuffer:BUFFER-FIELD(4)
      ichFileMode = ihnField:buffer-value
      .
  if iplgHasCreate
  then substring( ichFileMode, 1, 1 ) = "C".
  if iplgHasRead
  then substring( ichFileMode, 2, 1 ) = "R".
  if iplgHasUpdate
  then substring( ichFileMode, 3, 1 ) = "U".
  if iplgHasDelete
  then substring( ichFileMode, 4, 1 ) = "D".
  ihnField:buffer-value = ichFileMode.  
end procedure.
/* -------------------------------------------------------------------------- */
procedure spCreateConnectorTag:
  define input  parameter ipinConnectorID as integer    no-undo.
  define input  parameter ipchProperty as character  no-undo.
  define input  parameter ipchValue as character  no-undo.
  define output parameter opinConnectorTagID as integer    no-undo.
  
  define buffer ibfConnectorTag for Repository.t_ConnectorTag.
  define variable ihnBuffer as handle no-undo.
  define variable ihnField as handle no-undo.
  define variable ichNote as character no-undo.

  if length( ipchValue ) gt 255
  then do:
    ichNote = ipchValue.
    ipchValue = "<memo>".
  end.
         
  create ibfConnectorTag.
  assign
      linConnectorTagID  = linConnectorTagID + 1
      opinConnectorTagID = linConnectorTagID
      ibfConnectorTag.EA_GUID = sfGetEAGUID()
      ibfConnectorTag.PropertyID = linConnectorTagID
      ibfConnectorTag.ElementID = ipinConnectorID
      ibfConnectorTag.Property = ipchProperty
      ihnBuffer = buffer ibfConnectorTag:handle
      ihnfield = ihnbuffer:BUFFER-FIELD(4)
      ihnField:BUFFER-VALUE = ipchValue
      ibfConnectorTag.Notes = ichNote
      .
end procedure.
/* -------------------------------------------------------------------------- */
procedure spFindStereotype:
  define input parameter ipchStereotype as character no-undo.
  define input parameter ipchAppliesTo as character no-undo.
  define output parameter opchEAGUID as character no-undo.
  
  define buffer ibfStereotype for Repository.t_Stereotype.
  
  find ttStereotype
      where ttStereotype.chStereotype eq ipchStereotype
      and   ttStereotype.chAppliesTo  eq ipchAppliesTo
      no-error.
  if not available ( ttStereotype )
  then do:
    create ibfStereotype.
    assign
        ibfStereotype.EA_GUID = sfGetEAGUID()
        ibfStereotype.Stereotype = ipchStereotype
        ibfStereotype.AppliesTo  = ipchAppliesTo
        /* ibfStereotype.Description */
        ibfStereotype.MFEnabled = -1
        /* ibfStereotype.MFPath */
        /* ibfStereotype.MeteFile */
        /* ibfStereotype.Style */
        /* ibfStereotype..VisualType */
        .        
    create ttStereotype.
    assign
        ttStereotype.chStereotype = ipchStereotype
        ttStereotype.chAppliesTo  = ipchAppliesTo
        ttStereotype.chEAGUID = ibfStereotype.EA_GUID
        .
    lobExceptionStack:pushException( 981, 4, "Creating Stereotype = " + ipchStereotype + " to apply to " + ipchAppliesTo ).        
  end.
  opchEAGUID = ttStereotype.chEAGUID.
  
end procedure.
/* -------------------------------------------------------------------------- */
procedure spMakeStereotypeXref:
  define input parameter ipchClientType as character no-undo.
  define input parameter ipchStereotype as character no-undo.
  define input parameter ipchClientGUID as character no-undo.
  
  define variable ichStereotypeGUID as character no-undo.
  define variable ichType as character no-undo.
  
  define buffer ibfXref for Repository.t_xref.
  
  run spFindStereotype (
      input ipchStereotype,
      input ipchClientType,
      output ichStereotypeGUID
      ).
  if ichStereotypeGUID eq ?
  then do:
    lobExceptionStack:pushException( 982, 2, "Stereotype = " + ipchStereotype + " not found in spMakeStereotypeXref for Client Type = " + ipchClientType ).
    return.    
  end.
  
  ichType = sfGetXrefType( ipchClientType ).    
  
  create ibfXref.
  assign
      ibfXref.XrefID = sfGetEAGUID()
      ibfXref.Name = "Stereotypes"
      ibfXref.Type = ichType
      ibfXref.Visibility = "Public"
      /* ibfXref.Namespace */
      /* ibfXref.Requirement */
      /* ibfXref.Constraint */
      /* ibfXref.Behavior */
      /* ibfXref.Partition */
      ibfXref.Description = "@STEREO;Name=" + ipchStereotype + ";GUID=" + ichStereotypeGUID + ";@ENDSTEREO"
      ibfXref.Client = ipchClientGUID
      ibfXref.Supplier = "<none>"
      /* ibfXref.Link */
      .
  
end procedure.
/* -------------------------------------------------------------------------- */
procedure spChangeStereotypeXref:
  define input parameter ipchClientType as character no-undo.
  define input parameter ipchOldStereotype as character no-undo.
  define input parameter ipchNewStereotype as character no-undo.  
  define input parameter ipchClientGUID as character no-undo.
  
  define variable ichOldStereotypeGUID as character no-undo.
  define variable ichNewStereotypeGUID as character no-undo.
  define variable ichType as character no-undo.
  
  define buffer ibfXref for Repository.t_xref.
  
  run spFindStereotype (
      input ipchOldStereotype,
      input ipchClientType,
      output ichOldStereotypeGUID
      ).
  if ichOldStereotypeGUID eq ?
  then do:
    lobExceptionStack:pushException( 983, 2, "Stereotype = " + ipchOldStereotype + " not found in spChangeStereotypeXref for Client Type = " + ipchClientType ).
    return.    
  end.    
  run spFindStereotype (
      input ipchNewStereotype,
      input ipchClientType,
      output ichNewStereotypeGUID
      ).      
  if ichNewStereotypeGUID eq ?
  then do:
    lobExceptionStack:pushException( 984, 2, "Stereotype = " + ipchNewStereotype + " not found in spChangeStereotypeXref for Client Type = " + ipchClientType ).
    return.    
  end.    
  
  ichType = sfGetXrefType( ipchClientType ).
  
  for each ibfXref exclusive-lock
      where ibfXref.Client eq ipchClientGUID
      and   ibfXref.Type eq ichType
      :
    if ibfXref.Description matches "*" + ipchOldStereotype + "*"
    then do:
      ibfXref.Description = replace( ibfXref.Description, ipchOldStereotype, ipchNewStereotype ).    
      ibfXref.Description = replace( ibfXref.Description, ichOldStereotypeGUID, ichNewStereotypeGUID ).
    end.
  end.
  
end procedure.    
/* **************************** Functions *********************************** */
function sfCompileName returns character ( 
    input ipchProgramName as character 
    ):
  define variable iinDot as integer no-undo.
  
  iinDot = r-index( ipchProgramName, ".").
  if iinDot gt 0
  then return substring( ipchProgramName, 1, iinDot ) + "r".
  else return ipchProgramName + ".r".       
      
end function.
/* -------------------------------------------------------------------------- */
function sfGetEAGUID returns character :

  return "~{" + guid(generate-uuid) + "~}".

end function. 
/* -------------------------------------------------------------------------- */
function sfGetXrefType returns character (
    input ipchClientType as character
    ):

  define variable ichType as character no-undo.

  case ipchClientType:
    when "Component" then ichType = "element".
    when "Package" then ichType = "element".
    when "Class" then ichType = "element".
    when "AttribTable" then ichType = "element".
    when "Attribute" then ichType = "element".
    when "OpTable" then ichType = "element".
    when "Operation" then ichType = "element".
    when "Instantiation" then ichType = "connector".
    when "ControlFlow" then ichType = "connector".
    when "Dependency" then ichType = "connector".
    otherwise 
      do:
        lobExceptionStack:pushException( 991, 2, "spChangeStereotypeXref called with unknown Client Type = " + ipchClientType ).
        return ?.
      end.
  end case.    
  
  ichType = ichType + " property".
  return ichType.      
      
end function.      
/* ************************************************************************** */  
procedure spEAUtilitiesDestruct:
  delete object lobExceptionStack.
  lobExceptionStack = ?.
  empty temp-table ttObject.
  empty temp-table ttStereotype.
end procedure.       
/* ************************** End of Program ******************************** */